﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Windows.Forms;
using System.Runtime.InteropServices;

using ESRI.ArcGIS.Carto;
using ESRI.ArcGIS.Display;
using ESRI.ArcGIS.esriSystem;
using ESRI.ArcGIS.Geodatabase;
using ESRI.ArcGIS.Editor;
using ESRI.ArcGIS.Framework;

namespace JAMBWOW
{
    public static class AnnotationGenerator
    {
        #region Public Methods

        private const string ANNO_CANCELED_MESSAGE = "Annotation Generation Canceled";

        //This function takes a list of feature layers as input. It generates annotation layers by converting
        //the feature layer's labels to annotation. The annotation layers contain page-specific annotation features.
        //If pageIndex is specified, then page-specific features are generated for one page only, otherwise,
        //features are generated for all pages.
        public static bool GenerateAnnotation(List<ILayer> pLayersToConvert, int? pageIndex = null)
        {
            bool bOnePage;
            int iPageCount;
            int iStartPage;
            int iEndPage;
            int currentPageIndex;
            //Set parameters based on whether we are processing one page or all pages.
            if (pageIndex != null)
            {
                bOnePage = true;
                iPageCount = 1;
                iStartPage = (int)pageIndex;
                iEndPage = (int)pageIndex;
                currentPageIndex = (int)pageIndex;
            }
            else
            {
                bOnePage = false;
                iPageCount = DDP_Util.GetPageCount();
                iStartPage = 1;
                iEndPage = iPageCount;
                //Get the current page index (so that we can return to this page after completion)
                currentPageIndex = DDP_Util.GetCurrentKnownPage();
            }

            //Set this boolean to true when the lables are successfully converted to annotation.
            bool bSuccess = false; 
            
            //Display a progress window.
            ProgressWindow pProgressWindow = new ProgressWindow("Generating Annotation For Page");
            pProgressWindow.Show();

            //Create a progress report object.
            ProgressReport pProgressReport = new ProgressReport();
            pProgressReport.NumberOfItems = iPageCount;
            pProgressReport.CurrentItemNumber = 0;

            //Get the feature workspace containing the annotation layers.
            IFeatureWorkspace pFeatureWorkspace = AnnotationUtil.GetAnnoFeatureWorkspace();
            //Get the first data frame.
            IMap pMap = GenUtil.GetFirstDataFrame();

            try
            {               
                //Turn on labels for each layer before converting the labels to annotation.          
                foreach (ILayer pLayer in pLayersToConvert)
                {
                    IGeoFeatureLayer pGeoFeatureLayer = pLayer as IGeoFeatureLayer;
                    pGeoFeatureLayer.DisplayAnnotation = true;
                }

                //Loop through the DDP pages and Append Annotation for each page.
                //This will cause the annotation to be generated at a scale and rotation angle that is appropriate for each page.
                IConvertLabelsToAnnotation pConvertLabelsToAnnotation;
                
                for (int pageNum = iStartPage; pageNum <= iEndPage; pageNum++)
                {
                    //Update the progress window
                    pProgressReport.CurrentItemNumber = pageNum - iStartPage + 1;
                    bool Canceled = pProgressWindow.ProgressChanged(pProgressReport);
                    if (Canceled) throw new Exception(ANNO_CANCELED_MESSAGE);
                    System.Windows.Forms.Application.DoEvents(); //Allow the UI to refresh

                    //Navigate to the page (if processing multiple pages);
                    if (!bOnePage) DDP_Util.GoToPage(pageNum);

                    //Update the progress window after page navigation finishes.                
                    bool Canceled2 = pProgressWindow.ProgressChanged(pProgressReport);
                    if (Canceled2) throw new Exception(ANNO_CANCELED_MESSAGE);
                    System.Windows.Forms.Application.DoEvents(); //Allow the UI to refresh

                    //Initialize the cancel tracker object
                    ITrackCancel pTrackCancel = new CancelTrackerClass();
                    pTrackCancel.CancelOnClick = false;
                    pTrackCancel.CancelOnKeyPress = false;
                    //Assign the cancel tracker to the progress window.
                    pProgressWindow.CancelTracker = pTrackCancel;

                    //Generate annotation                   
                    pConvertLabelsToAnnotation = ConvertLabelsToAnno_MultiLayer(pMap, pLayersToConvert,
                        pFeatureWorkspace, pTrackCancel, pageNum);
                }

                //Add the layers to the table of contents if they are not already there.
                //Also turn off labels for the feature layers who's labels were converted to annotation.
                foreach (ILayer pLayer in pLayersToConvert)
                    AnnotationUtil.AddAnnotationToTOC(pLayer, pFeatureWorkspace);


                bSuccess = true;
                return bSuccess;
            }
            catch (Exception pException)
            {
                //Allow the UI to refresh (and give other processes a chance to finish thier work.).
                System.Windows.Forms.Application.DoEvents();

                if (pException.Message == ANNO_CANCELED_MESSAGE)
                    MessageBox.Show("Annotation Generation Canceled.", "JambWow - Annotation Generation Canceled");
                else
                    GenUtil.ShowError(pException);
                return bSuccess;
            }
            finally
            {
                if (!bOnePage) DDP_Util.GoToPage(currentPageIndex); //Return to the original page page
                pProgressWindow.Close(); //Close the progress Window.                
                if (bSuccess)
                {
                    if (!bOnePage)
                        MessageBox.Show("Finished Generating Annotation for " + iPageCount + " Page(s)", "JambWow - Finished Generating Annotation");
                    else
                        MessageBox.Show("Finished Generating Annotation for the current page", "JambWow - Finished Generating Annotation");
                }
            }
        }

        #endregion

        #region Private methods

        private static IConvertLabelsToAnnotation ConvertLabelsToAnno_MultiLayer(IMap pMap, List<ILayer> pLayers,
            IFeatureWorkspace pFeatureWorkspace, ITrackCancel pTrackCancel, int iPageIndex)
        {
            if (pLayers.Count == 0) return null;

            string sPageName = IndexLayerUtil.GetPageName(iPageIndex); //Get the page name.

            IConvertLabelsToAnnotation pConvertLabelsToAnnotation = new ConvertLabelsToAnnotationClass();
            bool generateUnplacedAnnotaiton = true;
            //Change global level options for the conversion by sending in different parameters to the next line.
            pConvertLabelsToAnnotation.Initialize(pMap, esriAnnotationStorageType.esriDatabaseAnnotation,
                esriLabelWhichFeatures.esriVisibleFeatures, generateUnplacedAnnotaiton, pTrackCancel, null);

            foreach (ILayer pLayer in pLayers)
            {                                               
                IGeoFeatureLayer pGeoFeatureLayer = pLayer as IGeoFeatureLayer;
                if (pGeoFeatureLayer != null)
                {
                    //Specify the parameters of the output annotation feature class here as well.                                    
                    string sAnnoLayerName = AnnotationUtil.GetAnnoLayerName(pLayer);                                     
                    bool bRequireSymbolID = false;                    
                    //If the annotation layer already exists, set append mode to true
                    IWorkspace2 pWorkspace = (IWorkspace2)pFeatureWorkspace;
                    bool bAppendMode = pWorkspace.get_NameExists(esriDatasetType.esriDTFeatureClass, sAnnoLayerName);
                    //Add the layer information to the converter object.
                    pConvertLabelsToAnnotation.AddFeatureLayer(pGeoFeatureLayer, sAnnoLayerName,
                        pFeatureWorkspace, null, false, bAppendMode, bRequireSymbolID, false, false, "");
                }
            }

            //Do the conversion (from text to annotation)
            pConvertLabelsToAnnotation.ConvertLabels();

            //Add and populate the page name field in the annotation layer.            
            foreach (ILayer pLayer in pLayers) //(Loop through each of the feature layers)
                AnnotationUtil.PopulatePageNameField(pLayer, pFeatureWorkspace, sPageName);
            
            //Return the conversion object.
            return pConvertLabelsToAnnotation;
        }

        #endregion
    }
}