﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Windows.Forms;
using System.Runtime.InteropServices;

using ESRI.ArcGIS.Carto;
using ESRI.ArcGIS.Display;
using ESRI.ArcGIS.esriSystem;
using ESRI.ArcGIS.Geodatabase;
using ESRI.ArcGIS.Editor;
using ESRI.ArcGIS.Framework;
using ESRI.ArcGIS.Catalog;

namespace JAMBWOW
{
    public static class AnnotationUtil
    {
        public const string ANNO_PAGE_NAME_FIELD = "PAGE_NAME";
        public const string ANNO_CONST = "_Anno";

        //public const string PAGE_NAME_ANNO_CONST = "_Anno_For_";
        //public const string ANNO_CONST = "_Anno_Page_";        
        //public const string GROUP_ANNO_CONST = "_Group_Anno";

        #region Public Methods

        //This function takes a feature layer name as input.
        //It returns the name of an annotation layer.
        public static string GetAnnoLayerName(ILayer pLayer)
        {
            //The annotation layer name is a concatenation of feature layer name, and a constant.
            string sName = pLayer.Name + ANNO_CONST;
            //Replace non-alpha-numeric characters with underscores
            string pattern = "[^a-zA-Z0-9_]+"; //regex pattern 
            sName = System.Text.RegularExpressions.Regex.Replace(sName, pattern, "_");
            return sName;
        }

        //This function returns true if an annotation layer in the table of contents
        //is associated with the feature layer (pBaseLayer) and follows the proper naming convention.
        //If true, it returns the associated annotation layer.
        public static bool HasAnnotationInTOC(ILayer pBaseLayer, out IGraphicsLayer pAnnoLayer)
        {
            pAnnoLayer = null;

            //Get an enumeration of annotation layers in the first data frame
            IMap pMap = GenUtil.GetFirstDataFrame();
            UID uid = new UIDClass();
            uid.Value = GenUtil.IANNOTATION_LAYER_UID;
            try
            {
                //Loop through the annotation layers
                IEnumLayer pEnumLayers = pMap.get_Layers(uid, true);
                pEnumLayers.Reset();
                ILayer pLayer = null;
                while ((pLayer = pEnumLayers.Next()) != null)
                {
                    IGraphicsLayer pGraphicsLayer = (IGraphicsLayer)pLayer;
                    ILayer pAssociatedLayer = pGraphicsLayer.AssociatedLayer;
                    if (pBaseLayer == pAssociatedLayer)
                    {
                        //If the page name matches the proper naming convention
                        if (pLayer.Name == GetAnnoLayerName(pBaseLayer))
                        {
                            pAnnoLayer = pGraphicsLayer;
                            return true;
                        }
                    }
                }
                //If we got this far without finding a match, return false 
                pAnnoLayer = null;
                return false;
            }
            catch (Exception e)
            {
                GenUtil.ShowError(e);
                return false;
            }
        }

        //This function takes a layer name, a feature workspace, and a page index as input.
        //It returns true if the feature workspace contains an associated annotation layer AND
        //the associated annotation layer contains at least one feature for the specified page index.
        public static bool HasAnnoFeaturesForPage(ILayer pLayer, IFeatureWorkspace pFeatureWorkspace, int pageIndex)
        {
            bool bMatch = false;
            IWorkspace2 pWorkspace = (IWorkspace2)pFeatureWorkspace;
            bool bExists = pWorkspace.get_NameExists(esriDatasetType.esriDTFeatureClass, GetAnnoLayerName(pLayer));
            if (!bExists) return false; //Return false if the annotation layer does not exist.

            //Get the page name from the page index
            string sPageName = IndexLayerUtil.GetPageName(pageIndex);
            //Get the annotation feature class
            IFeatureClass pAnnoFeatureClass = pFeatureWorkspace.OpenFeatureClass(GetAnnoLayerName(pLayer));
            //Querry for features that belong to the current page
            IFeatureCursor pFeatCur = null;
            try
            {
                ISpatialFilter pSpatialFilter = new SpatialFilterClass();
                pSpatialFilter.WhereClause = AnnotationUtil.ANNO_PAGE_NAME_FIELD + " = '" + sPageName + "'";
                pFeatCur = pAnnoFeatureClass.Search(pSpatialFilter, true);
                IFeature pFeature;
                while (((pFeature = pFeatCur.NextFeature()) != null) && !bMatch)
                    bMatch = true;
                return bMatch; //Return true if the query returns at least one feature.
            }
            catch (Exception pException)
            {
                GenUtil.ShowError(pException);
                return false;
            }
            finally
            {
                //Release the Cursor
                if (pFeatCur != null) System.Runtime.InteropServices.Marshal.FinalReleaseComObject(pFeatCur);                
            }
        }

        //This function takes a layer name and a feature workspace as input.
        //It returns true if the feature workspace contains an associated annotaion layer
        //that follows a pre-defined naming convention.
        public static bool HasAnnotationInDatabase(ILayer pLayer, IFeatureWorkspace pFeatureWorkspace)
        {
            IWorkspace2 pWorkspace = (IWorkspace2)pFeatureWorkspace;
            bool bExists = pWorkspace.get_NameExists(esriDatasetType.esriDTFeatureClass, GetAnnoLayerName(pLayer));
            return bExists;
        }

        //This function takes a feature layer as input.
        //It finds an annotation layer that is associated with the featur layer,
        //and it adds that annotation layer to the table of contents.
        public static void AddAnnotationToTOC(ILayer pLayer, IFeatureWorkspace pFeatureWorkspace)
        {
            //Get the feature workspace containing the annotation layers.            
            IWorkspace2 pWorkspace = (IWorkspace2)pFeatureWorkspace;
            
            //Get the annotation layer name (based on a naming convention)
            string sAnnotationLayerName = GetAnnoLayerName(pLayer);

            //Check if the annotation layer for this page really exists in the database.
            bool bExistsInDatabase = pWorkspace.get_NameExists(esriDatasetType.esriDTFeatureClass, sAnnotationLayerName);
            //Check if the annotation layer already exists in the TOC.
            IGraphicsLayer pTempLayer = null;
            bool bExistsInTOC = AnnotationUtil.HasAnnotationInTOC(pLayer, out pTempLayer);

            //If the annotation layer exists in the database, but not in the TOC, then add it to the TOC.
            if (bExistsInDatabase && !bExistsInTOC) 
            {       
                //Make a temporary annotation layer
                IFeatureClass pClass = pFeatureWorkspace.OpenFeatureClass(sAnnotationLayerName);
                IFeatureLayer pAnnoFeatLayer = (IFeatureLayer)new FDOGraphicsLayer();
                pAnnoFeatLayer.FeatureClass = pClass;
                pAnnoFeatLayer.Name = pClass.AliasName;
                //Associate the annotation layer to its feature layer
                IGraphicsLayer pGraphicsLayer = (IGraphicsLayer)pAnnoFeatLayer;
                pGraphicsLayer.AssociatedLayer = pLayer;

                //Generate a new annotation layer for the TOC (based on the temporary layer)
                //with a page definition query based on a layer file template,
                //but with other properties that come from the original layer.
                ILayer pNewAnnoLayer = ModifyAnnoLayerPropertiesFromLyrTemplate(pAnnoFeatLayer);

                //Add the new annotation layer to the map
                IMap pMap = GenUtil.GetFirstDataFrame();
                pMap.AddLayer(pNewAnnoLayer);
            }

            //Turn off labeling for the feature layer.
            IGeoFeatureLayer pGeoFeatureLayer = pLayer as IGeoFeatureLayer;
            pGeoFeatureLayer.DisplayAnnotation = false;
        }

        //This function takes a list of feature layers as input and generates associated page specific annotation.
        //It creates and populates annotation layers and adds them to the table of contents (TOC).
        //If bThisPageOnly is true, it generates annotation features for the current page only.
        public static bool ConvertLabelsToPageSpecificAnno(List<ILayer> pLayers, bool bThisPageOnly = false)
        {
            //If processing the current page only, get the current page index.
            int pageIndex = 0;
            if (bThisPageOnly) pageIndex = DDP_Util.GetCurrentKnownPage();

            bool bLayersProcessed = false;

            //Get the feature workspace containing the annotation layers.
            IFeatureWorkspace pFeatureWorkspace = GetAnnoFeatureWorkspace();
            IWorkspace2 pWorkspace = (IWorkspace2)pFeatureWorkspace;
            //Get the first data frame.
            IMap pMap = GenUtil.GetFirstDataFrame();
            
            //Loop through the list of feature layers.
            List<ILayer> pLayersToConvert = new List<ILayer>();
            foreach (ILayer pLayer in pLayers)            
            {
                //Determine if the layer already has a corresponding annotation layer
                //in the database and in the TOC.                
                IGraphicsLayer pAnnoLayer = null;                
                bool bHasAnnoInDB = AnnotationUtil.HasAnnotationInDatabase(pLayer, pFeatureWorkspace);
                bool bHasAnnoInTOC = AnnotationUtil.HasAnnotationInTOC(pLayer, out pAnnoLayer);

                //If the annotation layer already exists in the TOC,
                //but does not exist in the database,
                //remove the broken annotation layer from the TOC.                                
                if (!bHasAnnoInDB && bHasAnnoInTOC)
                    pMap.DeleteLayer((ILayer)pAnnoLayer);                        
                
                //If the annotation layer exists in the database but not in the map,
                //add it to the TOC and associate the annotation layers to its feature layer.
                if (bHasAnnoInDB && !bHasAnnoInTOC)
                {
                    bLayersProcessed = true;
                    AnnotationUtil.AddAnnotationToTOC(pLayer, pFeatureWorkspace);
                }

                if (bThisPageOnly) //If processing for the current page only
                {
                    //Determine if the layer contains annotation features for the current page.
                    bool bHasAnnoInDBForPage = AnnotationUtil.HasAnnoFeaturesForPage(pLayer, pFeatureWorkspace, pageIndex);
                    //If the annotation does not exist in the database, or if it exists in
                    //the database but does not have any features for the current page,
                    //add it to the list of layers to convert.
                    if (!bHasAnnoInDB || !bHasAnnoInDBForPage)
                        pLayersToConvert.Add(pLayer);
                }
                else //If processing for all pages
                {
                    //If the annotation layer does not already exists in the database,                
                    //add it to the list of layers to convert.
                    if (!bHasAnnoInDB) pLayersToConvert.Add(pLayer);
                }
            }

            if (pLayersToConvert.Count > 0)
            {
                bLayersProcessed = true;
                bool bSucccess;

                if (!bThisPageOnly) //Generate annotation for all pages
                    bSucccess = AnnotationGenerator.GenerateAnnotation(pLayersToConvert);
                else //Generate annotation for the current page only
                    bSucccess = AnnotationGenerator.GenerateAnnotation(pLayersToConvert, pageIndex);

                //If the annotation generator return false, it did not complete successfully.                
                if (!bSucccess)
                {
                    if (!bThisPageOnly) //If failed to generate annotation for multiple pages, delete the incomplete annotation layers.
                    {
                        MessageBox.Show("Annotaton conversion failed or was canceled. I will now delete any left over annotation.", "JambWow - Deleting Annotation");
                        try
                        {
                            AnnotationUtil.DeleteAnnotationLayers(pLayersToConvert);
                        }
                        catch (Exception pException)
                        {
                            GenUtil.ShowError(pException);
                        }
                    }
                }
            }

            if (bLayersProcessed)
            {
                //Refresh the table of contents
                GenUtil.GetMxDocument().UpdateContents();
                //Refresh the map to update the display.
                IActiveView pActiveView = pMap as IActiveView;
                pActiveView.Refresh();
            }
            return bLayersProcessed;
        }
        
        public static bool DeleteAnnotatioFeaturesThisPageOnly(List<ILayer> pFeatureLayers)
        {
            bool bLayersProcessed = false;

            //Get the feature workspace containing the annotation layers.
            IFeatureWorkspace pFeatureWorkspace = GetAnnoFeatureWorkspace();

            //Get the current page index
            int pageIndex = DDP_Util.GetCurrentKnownPage();
            //Get the page name from the page index
            string sPageName = IndexLayerUtil.GetPageName(pageIndex);

            //Get the first data frame.
            IMap pMap = GenUtil.GetFirstDataFrame();

            //Determine if layers have annotation in the database and/or in the table of contents
            //and determine if the layers contains features that correspond with the current page index.
            foreach (ILayer pFeatureLayer in pFeatureLayers)
            {
                bool bHasAnnoIDBForPage = AnnotationUtil.HasAnnoFeaturesForPage(pFeatureLayer, pFeatureWorkspace, pageIndex);

                //If annotation features exist in the database for this page, delete them.
                if (bHasAnnoIDBForPage)
                {
                    bLayersProcessed = true;
                    bool bDeleted = false;

                    //Get the annotation feature class
                    IFeatureClass pAnnoFeatureClass = pFeatureWorkspace.OpenFeatureClass(GetAnnoLayerName(pFeatureLayer));
                    //Querry for features that belong to the current page
                    IFeatureCursor pFeatCur = null;
                    try
                    {
                        ISpatialFilter pSpatialFilter = new SpatialFilterClass();
                        pSpatialFilter.WhereClause = AnnotationUtil.ANNO_PAGE_NAME_FIELD + " = '" + sPageName + "'";
                        pFeatCur = pAnnoFeatureClass.Search(pSpatialFilter, false);
                        IFeature pFeature;
                        while ((pFeature = pFeatCur.NextFeature()) != null)
                        {
                            pFeature.Delete();
                        }
                        bDeleted = true; //If the program got this far, deletion was successful.
                    }
                    catch (Exception pException)
                    {
                        GenUtil.ShowError(pException);
                    }
                    finally
                    {
                        if (pFeatCur != null) System.Runtime.InteropServices.Marshal.FinalReleaseComObject(pFeatCur); //Release the Cursor
                    }
                    if (bDeleted == false)
                    {
                        MessageBox.Show("Annotation for the '" + pFeatureLayer.Name + "' layer could not be deleted.", "JambWow - Failed to delete annotation features");
                        return bLayersProcessed;
                    }
                }
            }

            if (bLayersProcessed)
            {
                //*******************************************************
                //In a later version, we might turn on labels here.
                //As of now, that is not the desired behavior.
                //*******************************************************

                //Refresh the map to update the display.
                IActiveView pActiveView = pMap as IActiveView;
                pActiveView.Refresh();

                MessageBox.Show("Finished Deleting Annotation For The Current Page", "JambWow - Finished Deleting Annotation For The Current Page");
            }
            return bLayersProcessed;
        }

        public static bool DeleteAnnotationLayers(List<ILayer> pFeatureLayers)
        {
            bool bLayersProcessed = false;

            //Get the feature workspace containing the annotation layers.
            IFeatureWorkspace pFeatureWorkspace = GetAnnoFeatureWorkspace();
            
            //Get the first data frame.
            IMap pMap = GenUtil.GetFirstDataFrame();

            //Determine if layers have annotation in the database and/or in the table of contents.            
            foreach (ILayer pFeatureLayer in pFeatureLayers)
            {
                //Determine if the layer has a corresponding annotation layer
                //in the database and in the TOC.
                IGraphicsLayer pAnnoLayer;
                bool bHasAnnoInDB = AnnotationUtil.HasAnnotationInDatabase(pFeatureLayer, pFeatureWorkspace);
                bool bHasAnnoInTOC = AnnotationUtil.HasAnnotationInTOC(pFeatureLayer, out pAnnoLayer);

                //If the annotation exists in the TOC, remove the annotation layer from the TOC.                                                
                if (bHasAnnoInTOC)
                {
                    bLayersProcessed = true;
                    //Remove the annotation layer from the map
                    pMap.DeleteLayer((ILayer)pAnnoLayer);
                }

                //If the annotation exists in the database, delete it from the database
                if (bHasAnnoInDB)
                {
                    bLayersProcessed = true;
                    bool bDeleted = DeleteAnnotationLayerInDB(pFeatureLayer, pFeatureWorkspace);
                    if (bDeleted == false)
                        MessageBox.Show("Annotation for the '" + pFeatureLayer.Name + "' layer could not be deleted.", "JambWow - Failed to delete annotation layer");
                }
            }

            if (bLayersProcessed)
            {
                //Turn on labels for each layer
                foreach (ILayer pFeatureLayer in pFeatureLayers)
                {
                    IGeoFeatureLayer pGeoFeatureLayer = pFeatureLayer as IGeoFeatureLayer;
                    pGeoFeatureLayer.DisplayAnnotation = true;
                }

                //Refresh the table of contents
                GenUtil.GetMxDocument().UpdateContents();

                //Refresh the map to update the display.
                IActiveView pActiveView = pMap as IActiveView;
                pActiveView.Refresh();

                MessageBox.Show("Finished Deleting Annotation", "JambWow - Finished Deleting Annotation");
            }

            return bLayersProcessed;
        }

        public static ILayer ModifyAnnoLayerPropertiesFromLyrTemplate(ILayer pAnnoLayer)
        {
            //Get the annotation layer name
            string sAnnoLayerName = pAnnoLayer.Name;
            //Find the annotation layer's underlying annotation feature class.
            IFeatureLayer pFeatureLayer = (IFeatureLayer)pAnnoLayer;
            IFeatureClass pFeatureClass = pFeatureLayer.FeatureClass;
            //Find the annotation layer's assocaited feature layer.
            IGraphicsLayer pGraphicsLayer = (IGraphicsLayer)pAnnoLayer;
            ILayer pAssociatedLayer = pGraphicsLayer.AssociatedLayer;

            //Load a template annotation layer file with pre-existing page definition query.                    
            IGxLayer gxLayer = new GxLayerClass();
            IGxFile gxFile = (IGxFile)gxLayer;
            gxFile.Path = GenUtil.GetProgramPath() + @"\Templates\Anno_Template.lyr";
            ILayer pNewAnnoLayer = gxLayer.Layer;
            //Then overwrite the layer's properties with properties from the annotation layer.
            //Set the layer name.
            pNewAnnoLayer.Name = sAnnoLayerName;
            //Set the data source (annotation feature class)
            IFeatureLayer pNewFeatureLayer = (IFeatureLayer)pNewAnnoLayer;
            pNewFeatureLayer.FeatureClass = pFeatureClass;
            //Set the associated feature layer                    
            IGraphicsLayer pNewGraphicsLayer = (IGraphicsLayer)pNewAnnoLayer;
            pNewGraphicsLayer.AssociatedLayer = pAssociatedLayer;
            pNewGraphicsLayer.UseAssociatedLayerVisibility = true;
            //Change the expanded state of the new annotation layer
            ICompositeLayer2 pComposite = (ICompositeLayer2)pNewAnnoLayer;
            pComposite.Expanded = false;
            //Make the layer visible
            pNewAnnoLayer.Visible = true;

            //Return the new annotation layer
            return pNewAnnoLayer;
        }

        //Add a page name field (if it does not already exist)
        public static void CreatePageNameField(IFeatureClass pFC)
        {            
            int iFieldIndex = pFC.FindField(AnnotationUtil.ANNO_PAGE_NAME_FIELD);
            if (iFieldIndex == -1)
            {
                IField pField = new Field();
                IFieldEdit2 pFiedEdit = (IFieldEdit2)pField;
                pFiedEdit.Name_2 = AnnotationUtil.ANNO_PAGE_NAME_FIELD;
                pFiedEdit.Type_2 = esriFieldType.esriFieldTypeString;
                pFC.AddField(pField);
            }
        }

        public static void PopulatePageNameField(ILayer pLayer, IFeatureWorkspace pFeatureWorkspace, string sPageName)
        {
            string sAnnoLayerName = AnnotationUtil.GetAnnoLayerName(pLayer);
            IFeatureClass pFC = pFeatureWorkspace.OpenFeatureClass(sAnnoLayerName);

            //Add a page name field (if it does not already exist)            
            CreatePageNameField(pFC);

            // For any records with a null page name, replace with the current page name.
            IFeatureCursor pFeatCur = null;
            try
            {
                //Open the Annotation Feature Class for editing.  Create a Search Cursor
                ISpatialFilter pSpatialFilter = new SpatialFilterClass();
                pSpatialFilter.WhereClause = AnnotationUtil.ANNO_PAGE_NAME_FIELD + " is null";
                pFeatCur = pFC.Search(pSpatialFilter, false);
                IFeature pFeature;
                while ((pFeature = pFeatCur.NextFeature()) != null)
                {
                    //Get the index of the pagename field.
                    pFeature.set_Value(pFeatCur.Fields.FindField(AnnotationUtil.ANNO_PAGE_NAME_FIELD), sPageName);
                    pFeature.Store();
                }
            }
            catch (Exception pException)
            {
                GenUtil.ShowError(pException);
            }
            finally
            {
                //Release the Cursor
                if (pFeatCur != null) System.Runtime.InteropServices.Marshal.FinalReleaseComObject(pFeatCur);
            }
        }

        public static IFeatureWorkspace GetAnnoFeatureWorkspace()
        {
            //Get the paths to the MXD and File Geodatabase
            string sMXDPath = GenUtil.GetMapDocumentFileName();
            string sMXDFolder = Path.GetDirectoryName(sMXDPath);
            string sGDBName = Path.GetFileNameWithoutExtension(sMXDPath) + ".gdb";
            string sGDBPath = sMXDFolder + "\\" + sGDBName;

            //Create or open a file geodatabase at the same location as the map document.
            Type factoryType = Type.GetTypeFromProgID("esriDataSourcesGDB.FileGDBWorkspaceFactory");
            IWorkspaceFactory workspaceFactory = (IWorkspaceFactory)Activator.CreateInstance(factoryType);
            IWorkspace pWorkspace;
            if (!workspaceFactory.IsWorkspace(sGDBPath))
                workspaceFactory.Create(sMXDFolder, sGDBName, null, 0);

            //Open the the file geodatabase workspace.
            pWorkspace = workspaceFactory.OpenFromFile(sGDBPath, 0);

            IFeatureWorkspace pFeatureWorkspace = (IFeatureWorkspace)pWorkspace;
            return pFeatureWorkspace;
        }
                
        #endregion

        #region Private methods
       
        //This function deletes the annotation layer in the file geodatabase that is
        //associated with the feature layer pLayer.
        //It returns false if the annotation layer could not be deleted.
        private static bool DeleteAnnotationLayerInDB(ILayer pLayer, IFeatureWorkspace pFeatureWorkspace)
        {
            try
            {
                IWorkspace2 pWorkspace = (IWorkspace2)pFeatureWorkspace;
                string sAnnoLayerName = GetAnnoLayerName(pLayer);
                IFeatureClass pFeatureClass = pFeatureWorkspace.OpenFeatureClass(sAnnoLayerName);
                IDataset pDataset = (IDataset)pFeatureClass;
                if (pDataset.CanDelete() == false) return false;
                pDataset.Delete();
                return true;
            }
            catch (Exception pException)
            {
                GenUtil.ShowError(pException);
                return false;
            }
        }

        #endregion

        ////Prompt the user to replace any existing annotation layers in the database.
        //foreach (ILayer pLayer in pLayers)
        //{
        //    //Check if an annotation layer already exists for this layer.
        //    if (HasAnnotationInDatabase(pLayer))
        //    {
        //        //Delete the existing annotation layer before creating a new one.
        //        //Give the user the option to keep the existing annotation layer.
        //        DialogResult result = MessageBox.Show("Annotation already exists in the database for the '" + pLayer.Name + "' layer. Do you want to overwrite it?",
        //            "JambWow - Found Existing Annotation Layer", MessageBoxButtons.YesNo);
        //        if (result == System.Windows.Forms.DialogResult.Yes)
        //        {
        //            bool bDeleted = DeleteAnnotationLayerInDB(pLayer, pageCount, pFeatureWorkspace);
        //            if (bDeleted == false)
        //            {
        //                MessageBox.Show("Annotation for the '" + pLayer.Name + "' layer could not be deleted.", "JambWow - Failed to delete annotation layer");
        //                pLayers.Remove(pLayer);
        //            }
        //        }
        //        else
        //        {
        //            pLayers.Remove(pLayer);
        //        }
        //    }
        //}
    }
}
