﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Drawing;
using System.Runtime.InteropServices;
using ESRI.ArcGIS.Framework;
using ESRI.ArcGIS.ArcMapUI;
using ESRI.ArcGIS.Carto;
using ESRI.ArcGIS.Output;
using ESRI.ArcGIS.esriSystem;
using ESRI.ArcGIS.Geometry;
using ESRI.ArcGIS.Display;

namespace JAMBWOW
{
    public static class ExportUtil
    {
        /* GDI delegate to SystemParametersInfo function */
        [DllImport("user32.dll", SetLastError = true)]
        static extern bool SystemParametersInfo(uint uiAction, uint uiParam, ref int pvParam, uint fWinIni);

        /* constants used for user32 calls */
        const uint SPI_GETFONTSMOOTHING = 74;
        const uint SPI_SETFONTSMOOTHING = 75;
        const uint SPIF_UPDATEINIFILE = 0x1;

        public static void ExportPages(IExport docExport, ExportParameters pEP)
        {
            //Display a progress window.
            ProgressWindow pProgressWindow = new ProgressWindow("Exporting Page");
            pProgressWindow.Show();
            
            //Get starting page
            int startingPageNum = DDP_Util.GetCurrentKnownPage();
            int PageCount = DDP_Util.GetPageCount();
            //Get the base file name
            string sBaseFileName = pEP.BaseFileName;
            string sFileType = pEP.FileType.FileType;

            //Make a list of pages to export            
            List<int> PageNumbers = new List<int>();
            if (pEP.PageRangeType == PageRangeUtil.PageRangeTypes.PageList) //Make a list of select pages
            {
                //Parse the page numbers string
                string sPageNumbers = pEP.PageList;
                PageNumbers = PageRangeUtil.ParsePageNumbers(sPageNumbers);
            }
            else //Make a list of all pages
            {
                for (int pageNum = 1; pageNum <= PageCount; pageNum++)
                {
                    PageNumbers.Add(pageNum);
                }
            }

            //Remove disabled pages from PageNumbers list
            if (!pEP.OutputDisabledPages)
            {
                PageNumbers = PageRangeUtil.RemoveDisabledPages(PageNumbers);
            }

            //If Page Numbers list is empty, warn the user and exit.
            if (PageNumbers.Count == 0)
            {
                MessageBox.Show("All of the selected pages are disabled. No pages will be exported.", "JambWow");
                return;
            }

            ////Display a progress window.
            //ProgressWindow pProgressWindow = new ProgressWindow("Exporting Page");
            //pProgressWindow.Show();

            //Creat a progress report object.
            ProgressReport pProgressReport = new ProgressReport();
            pProgressReport.NumberOfItems = PageNumbers.Count;
            pProgressReport.CurrentItemNumber = 0;

            try
            {
                //Loop through and export each page
                for (int iIndex = 0; iIndex < PageNumbers.Count; iIndex++)
                {
                    int pageNum = PageNumbers[iIndex];

                    //Update the progress window
                    pProgressReport.CurrentItemNumber = iIndex + 1;
                    bool Canceled = pProgressWindow.ProgressChanged(pProgressReport);
                    if (Canceled)
                    {
                        DDP_Util.GoToPage(startingPageNum); //Return to the starting page
                        pProgressWindow.Close(); //Close the progress Window.
                        MessageBox.Show("Export Canceled", "JambWow");
                        return;
                    }
                    System.Windows.Forms.Application.DoEvents(); //Allow the UI to refresh

                    //Navigate to the page
                    DDP_Util.GoToPage(pageNum);

                    //Update the progress window after page navigation finishes.                
                    bool Canceled2 = pProgressWindow.ProgressChanged(pProgressReport);
                    if (Canceled2)
                    {
                        DDP_Util.GoToPage(startingPageNum); //Return to the starting page
                        pProgressWindow.Close(); //Close the progress Window.
                        MessageBox.Show("Export Canceled", "JambWow");
                        return;
                    }
                    System.Windows.Forms.Application.DoEvents(); //Allow the UI to refresh

                    //Initialize the cancel tracker object
                    ITrackCancel pTrackCancel = new CancelTrackerClass();
                    pTrackCancel.CancelOnClick = false;
                    pTrackCancel.CancelOnKeyPress = false;
                    //Assign the cancel tracker to the progress window.
                    pProgressWindow.CancelTracker = pTrackCancel;

                    //Export
                    ExportUtil.Export(docExport, sBaseFileName, sFileType, pTrackCancel);

                    //Remove the cancel tracker from the progress window.
                    pProgressWindow.CancelTracker = null;
                }

                DDP_Util.GoToPage(startingPageNum);//Return to the starting page
                pProgressWindow.Close(); //Close the progress Window.
                MessageBox.Show("Finished Exporting " + PageNumbers.Count + " Page(s)", "JambWow");
            }
            catch (Exception pException)
            {
                GenUtil.ShowError(pException);
                pProgressWindow.Close();
                DDP_Util.GoToPage(startingPageNum);//Return to the starting page
                MessageBox.Show("Export Failed", "JambWow");
            }
        }

        public static IExport GetExportClass(string ExportType)
        {
            // The Export*Class() type initializes a new export class of the desired type.
            IExport docExport;             
            if (ExportType == "EMF")
            {
                docExport = new ExportEMFClass();
            }
            else if (ExportType == "EPS")
            {
                docExport = new ExportPSClass();
            }
            else if (ExportType == "AI")
            {
                docExport = new ExportAIClass();
            }
            else if (ExportType == "PDF")
            {
                docExport = new ExportPDFClass();
            }
            else if (ExportType == "SVG")
            {
                docExport = new ExportSVGClass();
            }
            else if (ExportType == "BMP")
            {
                docExport = new ExportBMPClass();
            }
            else if (ExportType == "JPEG")
            {
                docExport = new ExportJPEGClass();
            }
            else if (ExportType == "PNG")
            {
                docExport = new ExportPNGClass();
            }
            else if (ExportType == "TIFF")
            {
                docExport = new ExportTIFFClass();
            }
            else if (ExportType == "GIF")
            {
                docExport = new ExportGIFClass();
            }
            else
            {
                MessageBox.Show("Unsupported export type " + ExportType + ", defaulting to EMF.", "JambWow");
                ExportType = "EMF";
                docExport = new ExportEMFClass();
            }
            return docExport;
        }

        public static void Export(IExport docExport, string sBaseFileName, string ExportType, ITrackCancel pTrackCancel)        
        {
            //**Output resolution and Resample Ratio are no longer hard-coded.**
            //long iOutputResolution = 300, long lResampleRatio = 1, 
            Boolean bClipToGraphicsExtent = false;

            //These were the original parameters that were used prior to this procedure being customized:

            /* EXPORT PARAMETER: (lResampleRatio) Output Image Quality of the export.  The value here will only be used if the export
             * object is a format that allows setting of Output Image Quality, i.e. a vector exporter.
             * The value assigned to ResampleRatio should be in the range 1 to 5.
             * 1 corresponds to "Best", 5 corresponds to "Fast"
             * EXPORT PARAMETER: (ExportType) a string which contains the export type to create.
             * EXPORT PARAMETER: (sOutputDir) a string which contains the directory to output to.
             * EXPORT PARAMETER: (bClipToGraphicsExtent) Assign True or False to determine if export image will be clipped to the graphic 
             * extent of layout elements.  This value is ignored for data view exports
             */

            string sCurrentPage = DDP_Util.GetCurrentKnownPage().ToString("0000");
            //MessageBox.Show("Export page: " + sCurrentPage, "JambWow");

            /* Exports the Layout View of the document to selected output format. */
            IActiveView docLayoutView = GenUtil.GetLayout();

            // using predefined static member            
            
            IPrintAndExport docPrintExport;
            
            string sNameRoot;
            bool bReenable = false;

            if (GetFontSmoothing())
            {
                /* font smoothing is on, disable it and set the flag to reenable it later. */
                bReenable = true;
                DisableFontSmoothing();
                if (GetFontSmoothing())
                {
                    //font smoothing is NOT successfully disabled, error out.
                    return;
                }
                //else font smoothing was successfully disabled.
            }

            //IExport docExport = GetExportClass(ExportType);
            
            docPrintExport = new PrintAndExportClass();

            //set the name root for the export
            sNameRoot = sBaseFileName + sCurrentPage;

            //set the export filename (which is the nameroot + the appropriate file extension)
            docExport.ExportFileName = sNameRoot + "." + docExport.Filter.Split('.')[1].Split('|')[0].Split(')')[0];

            //Output Image Quality of the export.  The value here will only be used if the export
            // object is a format that allows setting of Output Image Quality, i.e. a vector exporter.
            // The value assigned to ResampleRatio should be in the range 1 to 5.
            // 1 corresponds to "Best", 5 corresponds to "Fast"


            //*********************************************************************************************
            //* Do not set Resample Ratio programmatically.  It should be set in the Export Options pane. *
            //*********************************************************************************************
            //// check if export is vector or raster
            //if (docExport is IOutputRasterSettings)
            //{
            //    // for vector formats, assign the desired ResampleRatio to control drawing of raster layers at export time   
             //    IOutputRasterSettings RasterSettings = (IOutputRasterSettings)docExport;
                
            //    //RasterSettings.ResampleRatio = (int)lResampleRatio;
            //    //MessageBox.Show("Resample Ratio: " + RasterSettings.ResampleRatio.ToString(), "JambWow");

            //    // NOTE: for raster formats output quality of the DISPLAY is set to 1 for image export 
            //    // formats by default which is what should be used
            //}

            //MessageBox.Show("Export Resolution: " + docExport.Resolution, "JambWow");

            docPrintExport.Export(docLayoutView, docExport, docExport.Resolution, bClipToGraphicsExtent, pTrackCancel);

            //MessageBox.Show("Finished exporting " + sOutputDir + sNameRoot + "." + docExport.Filter.Split('.')[1].Split('|')[0].Split(')')[0] + ".", "JambWow - Export Active View Sample");

            if (bReenable)
            {
                /* reenable font smoothing if we disabled it before */
                EnableFontSmoothing();
                bReenable = false;
                if (!GetFontSmoothing())
                {
                    //error: cannot reenable font smoothing.
                    MessageBox.Show("Unable to reenable Font Smoothing", "JambWow");
                }
            }
        }

        private static void DisableFontSmoothing()
        {
            bool iResult;
            int pv = 0;

            /* call to systemparametersinfo to set the font smoothing value */
            iResult = SystemParametersInfo(SPI_SETFONTSMOOTHING, 0, ref pv, SPIF_UPDATEINIFILE);
        }

        private static void EnableFontSmoothing()
        {
            bool iResult;
            int pv = 0;

            /* call to systemparametersinfo to set the font smoothing value */
            iResult = SystemParametersInfo(SPI_SETFONTSMOOTHING, 1, ref pv, SPIF_UPDATEINIFILE);
        }

        private static Boolean GetFontSmoothing()
        {
            bool iResult;
            int pv = 0;

            /* call to systemparametersinfo to get the font smoothing value */
            iResult = SystemParametersInfo(SPI_GETFONTSMOOTHING, 0, ref pv, 0);

            if (pv > 0)
            {
                //pv > 0 means font smoothing is ON.
                return true;
            }
            else
            {
                //pv == 0 means font smoothing is OFF.
                return false;
            }
        }
    }
}
