﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using ESRI.ArcGIS.Carto;
using ESRI.ArcGIS.esriSystem;
using ESRI.ArcGIS.Framework;
using ESRI.ArcGIS.ArcMapUI;
using ESRI.ArcGIS.Geodatabase;

namespace JAMBWOW
{
    public partial class FormConvertLabelsToAnnotation : Form
    {
        private static FormConvertLabelsToAnnotation m_AnnoForm = null;
        private int? m_PageIndex = null;

        public static FormConvertLabelsToAnnotation ShowAnnoForm(int? pageIndex = null)
        {
            try
            {
                if (m_AnnoForm == null)
                {
                    m_AnnoForm = new FormConvertLabelsToAnnotation(pageIndex);
                    m_AnnoForm.Show(GenUtil.GetAppWinRef());
                    return m_AnnoForm;
                }
                else
                {
                    MessageBox.Show("The Convert Labels To Annotation dialog is already open.", "JambWow");
                    return null;
                }
            }
            catch (Exception e)
            {
                GenUtil.ShowError(e);
                return null;
            }
        }

        public static void Close_Form()
        {
            if (m_AnnoForm != null) m_AnnoForm.Close();                
        }

        //Constructor
        private FormConvertLabelsToAnnotation(int? pageIndex = null)
        {
            InitializeComponent();
            //Set a page index if we are converting lables to anno for one page only.
            m_PageIndex = pageIndex;
            //Set the window title
            if (m_PageIndex == null)
                this.Text = "Convert Labels To Annotation (All Pages)";
            else
                this.Text = "Convert Labels To Annotation (Page " + m_PageIndex.ToString() + ")";
            //Initialize the list of layers
            InitializeLayerList();           
        }

        #region GUI Functions

        private void FormConvertLabelsToAnnotation_FormClosing(object sender, FormClosingEventArgs e)
        {
            m_AnnoForm = null;
        }

        private void cmdCancel_Click(object sender, EventArgs e)
        {
            try
            {
                m_AnnoForm.Close();
            }
            catch (Exception pException)
            {
                GenUtil.ShowError(pException);
            }
        }

        private void cmdOK_Click(object sender, EventArgs e)
        {
            try
            {
                ConvertLabelsToAnnotation();
            }
            catch (Exception pException)
            {
                GenUtil.ShowError(pException);
            }
        }

        #endregion

        #region Non-GUI Functions

        private void InitializeLayerList()
        {
            //Set the mouse cursor to an hourglass until this method finishes.
            IMouseCursor appCursor = new MouseCursorClass();
            appCursor.SetCursor(2);

            //Display a progress window.
            ProgressWindow pProgressWindow = new ProgressWindow("Building List of Layers...");            
            //Create a progress report object.
            ProgressReport pProgressReport = new ProgressReport();
            pProgressReport.NumberOfItems = 1;
            pProgressReport.CurrentItemNumber = 1;
            //Update the progress window
            pProgressWindow.ProgressChanged(pProgressReport, true);
            pProgressWindow.Show();
            System.Windows.Forms.Application.DoEvents(); //Allow the UI to refresh

            //Get an enumeration of feature layers in the first data frame
            IMap pMap = GenUtil.GetFirstDataFrame();
            UID uid = new UIDClass();
            uid.Value = GenUtil.IFEATURE_LAYER_UID;
            try
            {
                //Get the page count (which is not used if we are converting to annotation for a single page.)
                int iPageCount = 0;
                if (m_PageIndex == null) iPageCount = DDP_Util.GetPageCount();

                //Get the annotation feature workspace
                IFeatureWorkspace pFeatureWorkspace = AnnotationUtil.GetAnnoFeatureWorkspace();

                lstLayers.Items.Clear();
                lstLayers.Sorted = false;
                IEnumLayer pEnumLayers = pMap.get_Layers(uid, true);

                //Loop through the layers
                pEnumLayers.Reset();
                ILayer pLayer = null;
                while ((pLayer = pEnumLayers.Next()) != null)
                {
                    if (!(pLayer is IAnnotationLayer) )
                    {
                        IGraphicsLayer pAnnoLayer;
                        string sLayerText = pLayer.Name;
                        bool bHasAnnoInTOC = AnnotationUtil.HasAnnotationInTOC(pLayer, out pAnnoLayer);
                        bool bHasAnnoInDB = AnnotationUtil.HasAnnotationInDatabase(pLayer, pFeatureWorkspace);                                                
                        if (m_PageIndex == null) //Generating Annotation All Pages
                        {
                            if (!(bHasAnnoInDB && bHasAnnoInTOC))
                            {
                                if (bHasAnnoInDB) sLayerText = @"[Already has annotation in database] - " + sLayerText;
                                if (bHasAnnoInTOC) sLayerText = @"[Already has annotation in TOC] - " + sLayerText;
                                LayerWrapper LW = new LayerWrapper(pLayer, sLayerText);
                                lstLayers.Items.Add(LW);
                            }
                        }
                        else //Generating Annotation For the Current Page Only
                        {
                            bool bHasAnnoInDBForPage = AnnotationUtil.HasAnnoFeaturesForPage(pLayer, pFeatureWorkspace, (int)m_PageIndex);
                            if (!(bHasAnnoInDBForPage && bHasAnnoInTOC))
                            {                                
                                if (bHasAnnoInDBForPage) sLayerText = @"[Already has annotation in database] - " + sLayerText;
                                LayerWrapper LW = new LayerWrapper(pLayer, sLayerText);
                                lstLayers.Items.Add(LW);
                            }
                        }
                    }              
                }
                lstLayers.Sorted = true;
            }
            catch (Exception e)
            {
                GenUtil.ShowError(e);
            }
            finally
            {
                pProgressWindow.Close(); //Close the progress Window. 
            }
        }

        private void ConvertLabelsToAnnotation()
        {
            try
            {                
                if (lstLayers.SelectedItems.Count == 0)
                {
                    MessageBox.Show("You must select at least one layer.", "JambWow");
                    return;
                }

                //Build a list of layers
                List<ILayer> pLayers = new List<ILayer>(); //Complete list of layers.
                foreach (object oSelectedItem in lstLayers.SelectedItems)
                {
                    LayerWrapper LW = (LayerWrapper)oSelectedItem;
                    ILayer pLayer = LW.Layer;
                    string sLayerName = pLayer.Name;
                    if (pLayer.Valid)
                    {
                        pLayers.Add(LW.Layer);
                    }
                    else
                    {
                        MessageBox.Show("The '" + sLayerName + "' layer is not valid. Repair its data source and try again. I cannot convert its labels to annotation.", "JambWow");
                        return;
                    }
                }
                m_AnnoForm.Hide();
                if (m_PageIndex == null)
                    AnnotationUtil.ConvertLabelsToPageSpecificAnno(pLayers);
                else
                    AnnotationUtil.ConvertLabelsToPageSpecificAnno(pLayers, true);
                m_AnnoForm.Close(); //Close the form.
            }
            catch (Exception e)
            {
                GenUtil.ShowError(e);
            }
        }

        #endregion
    }
}
