﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using ESRI.ArcGIS.Carto;
using ESRI.ArcGIS.esriSystem;
using ESRI.ArcGIS.Framework;
using ESRI.ArcGIS.ArcMapUI;
using ESRI.ArcGIS.Geodatabase;

namespace JAMBWOW
{
    public partial class FormDeleteAnnotation : Form
    {
        private static FormDeleteAnnotation m_DeleteAnnoForm = null;
        private int? m_PageIndex = null;

        public static FormDeleteAnnotation ShowDeleteAnnoForm(int? pageIndex = null)
        {
            try
            {
                if (m_DeleteAnnoForm == null)
                {
                    m_DeleteAnnoForm = new FormDeleteAnnotation(pageIndex);
                    m_DeleteAnnoForm.Show(GenUtil.GetAppWinRef());
                    return m_DeleteAnnoForm;
                }
                else
                {
                    MessageBox.Show("The Delete Annotation dialog is already open.", "JambWow");
                    return null;
                }
            }
            catch (Exception e)
            {
                GenUtil.ShowError(e);
                return null;
            }
        }

        public static void Close_Form()
        {
            if (m_DeleteAnnoForm != null) m_DeleteAnnoForm.Close();
        }

        //Constructor
        private FormDeleteAnnotation(int? pageIndex = null)
        {
            InitializeComponent();
            //Set a page index if we are deleting anno for one page only.
            m_PageIndex = pageIndex;
            //Set the window title
            if (m_PageIndex == null)
                this.Text = "Delete Annotation (All Pages)";
            else
                this.Text = "Delete Annotation (Page " + m_PageIndex.ToString() + ")";
            //Initialize the list of layers
            InitializeLayerList();
        }

        #region GUI Functions

        private void FormDeleteAnnotation_FormClosing(object sender, FormClosingEventArgs e)
        {
            m_DeleteAnnoForm = null;
        }

        private void cmdCancel_Click(object sender, EventArgs e)
        {
            try
            {
                m_DeleteAnnoForm.Close();
            }
            catch (Exception pException)
            {
                GenUtil.ShowError(pException);
            }
        }

        private void cmdOK_Click(object sender, EventArgs e)
        {
            try
            {
                DeleteAnnotation();
            }
            catch (Exception pException)
            {
                GenUtil.ShowError(pException);
            }
        }

        #endregion

        #region Non-GUI Functions

        public void InitializeLayerList()
        {
            //Set the mouse cursor to an hourglass until this method finishes.
            IMouseCursor appCursor = new MouseCursorClass();
            appCursor.SetCursor(2);

            //Display a progress window.
            ProgressWindow pProgressWindow = new ProgressWindow("Building List of Layers...");
            //Create a progress report object.
            ProgressReport pProgressReport = new ProgressReport();
            pProgressReport.NumberOfItems = 1;
            pProgressReport.CurrentItemNumber = 1;            
            //Update the progress window
            pProgressWindow.ProgressChanged(pProgressReport, true);
            pProgressWindow.Show();
            System.Windows.Forms.Application.DoEvents(); //Allow the UI to refresh

            //Get an enumeration of feature layers in the first data frame
            IMap pMap = GenUtil.GetFirstDataFrame();
            UID uid = new UIDClass();
            uid.Value = GenUtil.IFEATURE_LAYER_UID;
            try
            {
                //Get the page count (which is not used if we are deleting annotation for a single page.)
                int iPageCount = 0;
                if (m_PageIndex == null) iPageCount = DDP_Util.GetPageCount();

                //Get the annotation feature workspace
                IFeatureWorkspace pFeatureWorkspace = AnnotationUtil.GetAnnoFeatureWorkspace();

                lstLayers.Items.Clear();
                lstLayers.Sorted = false;
                IEnumLayer pEnumLayers = pMap.get_Layers(uid, true);

                //Loop through the layers
                pEnumLayers.Reset();
                ILayer pLayer = null;
                while ((pLayer = pEnumLayers.Next()) != null)
                {
                    if (!(pLayer is IAnnotationLayer) )
                    {
                        IGraphicsLayer pAnnoLayer;
                        string sLayerText = pLayer.Name;
                        bool bHasAnnoInTOC = AnnotationUtil.HasAnnotationInTOC(pLayer, out pAnnoLayer);
                        bool bHasAnnoInDB = AnnotationUtil.HasAnnotationInDatabase(pLayer, pFeatureWorkspace);                        
                        if (m_PageIndex == null) //Deleting Annotation for All Pages
                        {
                            if (bHasAnnoInDB || bHasAnnoInTOC)
                            {
                                if (!bHasAnnoInDB) sLayerText = @"[Does not have annotation in database] - " + sLayerText;
                                if (!bHasAnnoInTOC) sLayerText = @"[Does not have has annotation in TOC] - " + sLayerText;
                                LayerWrapper LW = new LayerWrapper(pLayer, sLayerText);
                                lstLayers.Items.Add(LW);
                            }                                                        
                        }
                        else //Deleting Annotation For Current Page Only
                        {                            
                            bool bHasAnnoInDBForPage = AnnotationUtil.HasAnnoFeaturesForPage(pLayer, pFeatureWorkspace, (int)m_PageIndex);                            
                            if (bHasAnnoInDBForPage)
                            {                                
                                if (!bHasAnnoInTOC) sLayerText = @"[Does not have has annotation in TOC] - " + sLayerText;
                                LayerWrapper LW = new LayerWrapper(pLayer, sLayerText);
                                lstLayers.Items.Add(LW);
                            }
                        }
                    }              
                }
                lstLayers.Sorted = true;
            }
            catch (Exception e)
            {
                GenUtil.ShowError(e);
            }
            finally
            {
                pProgressWindow.Close(); //Close the progress Window.                
            }
        }

        private void DeleteAnnotation()
        {
            if (lstLayers.SelectedItems.Count == 0)
            {
                MessageBox.Show("You must select at least one layer.", "JambWow");
                return;
            }

            //Set the mouse cursor to an hourglass until this method finishes.
            IMouseCursor appCursor = new MouseCursorClass();
            appCursor.SetCursor(2);

            //Display a progress window.
            ProgressWindow pProgressWindow = new ProgressWindow("Deleting Annotation...");
            //Create a progress report object.
            ProgressReport pProgressReport = new ProgressReport();
            pProgressReport.NumberOfItems = 1;
            pProgressReport.CurrentItemNumber = 1;
            //Update the progress window
            pProgressWindow.ProgressChanged(pProgressReport, true);
            pProgressWindow.Show();
            System.Windows.Forms.Application.DoEvents(); //Allow the UI to refresh

            try
            {
                //Build a list of layers
                List<ILayer> pLayers = new List<ILayer>(); //Complete list of layers.
                foreach (object oSelectedItem in lstLayers.SelectedItems)
                {
                    LayerWrapper LW = (LayerWrapper)oSelectedItem;
                    pLayers.Add(LW.Layer);
                }
                m_DeleteAnnoForm.Hide();
                if (m_PageIndex == null)
                    AnnotationUtil.DeleteAnnotationLayers(pLayers);
                else
                    AnnotationUtil.DeleteAnnotatioFeaturesThisPageOnly(pLayers);
                m_DeleteAnnoForm.Close(); //Close the form.
            }
            catch (Exception e)
            {
                GenUtil.ShowError(e);
            }
            finally
            {
                pProgressWindow.Close(); //Close the progress Window.                
            }
        }

        #endregion
    }
}
