﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using ESRI.ArcGIS.Geodatabase;
using ESRI.ArcGIS.Framework;
using ESRI.ArcGIS.ArcMapUI;
using ESRI.ArcGIS.Carto;
using ESRI.ArcGIS.esriSystem;

namespace JAMBWOW
{
    public static class IndexLayerUtil
    {        
        public const string INDEX_LAYER_NAME = "INDEX_GRID";
        public const string NAME_FIELD = "SMAP_NAME";
        public const string SORT_FIELD = "SMAP_NUM";
        public const string PAGE_NUMBER_FIELD = "SMAP_NUM";
        public const string FROM_STATION_FIELD = "FROM_STAT";
        public const string TO_STATION_FIELD = "TO_STAT";
        public const string MINIMUM_Y_FIELD = "MIN_Y";
        public const string Y_RANGE_FIELD = "Y_RANGE";
        public const string ROTATION_FIELD = "SMAP_ANGLE";
        
        //public const string SPATIAL_REFERENCE = "";
        //public const string SCALE_FIELD = "SMAP_SCALE";
        //public const string FROM_TO_TEX_FIELD = "ST_FROM_TO";
        //public const string IN_GRID_FIELD = "IN_GRID";                
        //public const string SHAPE_FIELD = "SHAPE";
        //public const string SORT_ORDER = "ASCENDING";
        //public const string DATA_FRAME_NAME = "Layers";

        //PUBLIC const int STARTING_PAGE_NUMBER = 1;

        //Get a reference to the Index Layer
        private static IFeatureLayer GetIndexLayer()
        {
            IMap pMap = GenUtil.GetFirstDataFrame();

            //Get an enumeration of feature layers in the data frame
            UID uid = new UIDClass();
            uid.Value = GenUtil.IFEATURE_LAYER_UID;
            try
            {
                IEnumLayer pEnumLayers = pMap.get_Layers(uid, true);

                //Loop through the layers
                pEnumLayers.Reset();                
                int iMatchesFound = 0;
                ILayer pMatchingLayer = null;
                ILayer pLayer = pEnumLayers.Next();
                while (!(pLayer == null))
                {
                    if (pLayer.Name.ToUpper() == INDEX_LAYER_NAME)
                    {
                        iMatchesFound++;
                        pMatchingLayer = pLayer;
                    }
                    pLayer = pEnumLayers.Next();
                }
                if (iMatchesFound == 0)
                {
                    MessageBox.Show("The first data frame is required to contain an index layer named '" + INDEX_LAYER_NAME + "'.", "JambWow - Missing index layer");
                    return null;
                }
                else if (iMatchesFound > 1)
                {
                    MessageBox.Show("The first data frame contains " + iMatchesFound.ToString() + " index layers named '" + INDEX_LAYER_NAME + "'. Only one is allowed.", "JambWow - Duplicate index layers");
                    return null;
                }
                return (IFeatureLayer)pMatchingLayer;
            }
            catch (Exception e)
            {
                GenUtil.ShowError(e);
                return null;
            }
        }

        //This function takes a page name as input and returns the page index of that page.
        //If the page is not found, it returns -1
        public static int LookupPageIndexFromPageName(string PageName)
        {
            //Get an array of page names
            string[] ArrPageNames = GetPageNames();
            if (ArrPageNames == null) return -1;
            if (ArrPageNames.GetLength(0) == 0) return -1;

            for (int n = 1; n <= ArrPageNames.GetLength(0); n++)
            {
                string sPageName = ArrPageNames[n - 1];
                if (sPageName == PageName) return n;
            }
            return -1; //If no match was found, return -1.
        }

        public static string GetRotation(int pageIndex)
        {
            return GetAttributeValue(ROTATION_FIELD, pageIndex, false);
        }

        public static string GetMinimumY(int pageIndex)
        {
            return GetAttributeValue(MINIMUM_Y_FIELD, pageIndex, false);
        }

        public static string GetYRange(int pageIndex)
        {
            return GetAttributeValue(Y_RANGE_FIELD, pageIndex, false);
        }

        public static string GetPageName(int pageIndex)
        {
            return GetAttributeValue(NAME_FIELD, pageIndex);
        }
        
        public static string GetFromStation(int pageIndex)
        {
            return GetAttributeValue(FROM_STATION_FIELD, pageIndex);
        }
        
        public static string GetToStation(int pageIndex)
        {
            return GetAttributeValue(TO_STATION_FIELD, pageIndex);
        }

        public static string GetAttributeValue(string AttributeName, int pageIndex, bool required = true)
        {
            try
            {
                //Get page count.
                int pageCount = DDP_Util.GetPageCount();                 
                //Get an array of page items
                string[] ArrAttributeValues = GetAttributeValues(AttributeName, required);
                if (ArrAttributeValues == null) return null;
                if (ArrAttributeValues.GetLength(0) == 0) return null;
                if (ArrAttributeValues.GetLength(0) != pageCount)
                {
                    MessageBox.Show("The number of pages in your index layer does not match the number of pages in Data Driven Pages. Refresh DDP and try again.", "JambWow");
                    return null;
                }

                //Validate the pageIndex
                if (pageIndex < 1)
                {
                    MessageBox.Show("Page index must be a positive integer.", "JambWow");
                    return null;
                }
                if (pageIndex > pageCount)
                {
                    MessageBox.Show("Page index cannot be greater than the number of pages.", "JambWow");
                    return null;
                }

                //Return the value at the specified page index.
                //Subtract 1 from the index because array indexes start at zero while DDP indexes start at one.
                return ArrAttributeValues[pageIndex - 1];
            }
            catch (Exception pException)
            {
                GenUtil.ShowError(pException);
                return null;
            }
        }

        public static string[] GetPageNumbers()
        {
            return GetAttributeValues(PAGE_NUMBER_FIELD);
        }

        public static string[] GetPageNames()
        {
            return GetAttributeValues(NAME_FIELD);
        }

        public static string[] GetAttributeValues(string AttributeName, bool required = true)
        {
            //Get the index layer
            IFeatureLayer pFeatureLayer = IndexLayerUtil.GetIndexLayer();
            if (pFeatureLayer == null)
            {
                MessageBox.Show("Failed to get feature layer.", "JambWow");
                return null;
            }

            //Make a query filter            
            IQueryFilter pQueryFilter = new QueryFilterClass();

            //Loop through the features and pull out the attribute value
            IFeatureCursor pFeatureCursor = pFeatureLayer.Search(pQueryFilter, false);
            int AttributeIndex = pFeatureCursor.FindField(AttributeName);
            int SortFieldIndex = pFeatureCursor.FindField(SORT_FIELD);

            if (AttributeIndex == -1)
            {
                if (required)
                    MessageBox.Show("The index layer '" + INDEX_LAYER_NAME + "' does not contain field called '" + AttributeName + "'. This field is required.", "JambWow - Missing AttributeName Field");
                return null;
            }

            if (SortFieldIndex == -1)
            {
                MessageBox.Show("The index layer '" + INDEX_LAYER_NAME + "' does not contain a name field called '" + SORT_FIELD + "'. This field is required.", "JambWow - Missing Sort Field");
                return null;
            }

            SortedDictionary<object, string> AttributeValues = new SortedDictionary<object, string>();

            IFeature pFeature = pFeatureCursor.NextFeature();
            while (pFeature != null)
            {
                string sAttribueValue = pFeature.get_Value(AttributeIndex).ToString();
                object oSortFieldValue = pFeature.get_Value(SortFieldIndex);
                if (!(oSortFieldValue is IComparable)) throw new ExceptionNullPageNumber("The page number field is not allowed to contain null values.");
                try
                {
                    AttributeValues.Add(oSortFieldValue, sAttribueValue);
                }
                catch
                {
                    throw new ExceptionDuplicatePageNumber("The page number field is not allowed to contain duplicate values.");
                }
                pFeature = pFeatureCursor.NextFeature();
            }

            if (AttributeValues.Count == 0) return null;

            //Copy the values to an ordered array of values.
            string[] ArrAttributeValues = new string[AttributeValues.Count];
            AttributeValues.Values.CopyTo(ArrAttributeValues, 0);

            //Return an array of values
            return ArrAttributeValues;
        }
    }
}
