﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using ESRI.ArcGIS.Carto;
using ESRI.ArcGIS.Geometry;
using ESRI.ArcGIS.esriSystem;
using ESRI.ArcGIS.Geodatabase;
using ESRI.ArcGIS.Display;
using ESRI.ArcGIS.ArcMapUI;


namespace JAMBWOW
{
    public static class JambMapUpgrader
    {
        //Annotation Constants
        private const string OLD_ANNO_CONST = "_Anno_Page_";
        private const string ANNO_CONVERSION_CANCELED_MESSAGE = "Annotation Conversion Canceled";
        //Hidden layout element constants
        private const double HIDE_LOCATION_X = -214748.3648;
        private const double HIDE_LOCATION_Y = -214748.3648;
        private const string TAG_CONST_THISPAGEONLY = "DSMAPBOOK - THISPAGEONLY";
        private const string TAG_CONST_ALLPAGESBUTTHIS = "DSMAPBOOK - ALLPAGESBUTTHIS";
        private const string PREFIX_THISPAGEONLY =    "Show only on page: ";
        private const string PREFIX_ALLPAGESBUTTHIS = "Show on all pages except: ";
        //Other layout element tagging constants
        private const string TAG_CONST_DATE = "DSMAPBOOK - DATE";
        private const string TAG_CONST_DATE2 = "date";
        private const string TAG_CONST_TITLE = "DSMAPBOOK - TITLE";
        private const string TAG_CONST_PAGENUMBER = "DSMAPBOOK - PAGENUMBER";
        private const string TAG_CONST_EXTRAITEM = "DSMAPBOOK - EXTRAITEM";
        private const string TAG_CONST_DATAGRAPH = "DSMAPBOOK - DATAGRAPH";       

        //This procedure copies annotation from map annotation groups into geodatabase annotation layers.
        public static void UpgradeMapAnnotationToGeodatabase()
        {
            // Create the annotation layer factory.
            IAnnotationLayerFactory annotationLayerFactory = new FDOGraphicsLayerFactoryClass();
            //Get the file geodatabase worksapce. Create the file geodatabase if it does not already exist.
            IFeatureWorkspace pFeatureWorkspace = AnnotationUtil.GetAnnoFeatureWorkspace();
            //Get the first data frame.
            IMap pMap = GenUtil.GetFirstDataFrame();
            //Get a clone of the spatial reference of the map.
            IClone pSRClone = ((IClone)pMap.SpatialReference).Clone();
            ISpatialReference pSR = (ISpatialReference)pSRClone;

            //Get a count of the number of annotation layers in the map.
            ICompositeLayer pCompositeLayer = (ICompositeLayer)pMap.BasicGraphicsLayer;
            int iGraphicLayerCount = pCompositeLayer.Count;

            //Display a progress window.
            ProgressWindow pProgressWindow = new ProgressWindow("Moving Annotation to Geodatabase");
            pProgressWindow.Show();

            //Create a progress report object.
            int iCount = 0;
            ProgressReport pProgressReport = new ProgressReport();
            pProgressReport.NumberOfItems = iGraphicLayerCount;
            pProgressReport.CurrentItemNumber = iCount;

            try
            {
                //Make a list of map annotation layers to delete
                List<string> MapAnnoLayersToDelete = new List<string>();

                //Loop through each map annotation group.
                for (int i = 0; i < iGraphicLayerCount; i++)
                {
                    //Update the progress window
                    iCount++;
                    pProgressReport.CurrentItemNumber = iCount;
                    bool Canceled = pProgressWindow.ProgressChanged(pProgressReport);
                    if (Canceled) throw new Exception(ANNO_CONVERSION_CANCELED_MESSAGE);
                    System.Windows.Forms.Application.DoEvents(); //Allow the UI to refresh

                    //Get the map annotation group
                    ILayer pMapAnnoLayer = pCompositeLayer.get_Layer(i);

                    //Get the associated feature layer.
                    IGraphicsLayer pMapAnnoGraphicsLayer = (IGraphicsLayer)pMapAnnoLayer;
                    ILayer pAssociatedFeatureLayer = pMapAnnoGraphicsLayer.AssociatedLayer;
                    if (pAssociatedFeatureLayer != null)
                    {
                        //Get the name of the map annotation group.
                        string sMapAnnoLayerName = pMapAnnoLayer.Name;

                        //Get the page index that appears after OLD_ANNO_CONST.
                        string[] stringSeperators = new string[] { OLD_ANNO_CONST };
                        string[] Tokens = sMapAnnoLayerName.Split(stringSeperators, StringSplitOptions.None);
                        string sPageIndex = Tokens.Last();

                        int iPageIndex;
                        bool isNum = int.TryParse(sPageIndex, out iPageIndex);
                        if (isNum) //If the annotation layer name contains a page index.
                        {
                            //Get the page name
                            string sPageName = IndexLayerUtil.GetPageName(iPageIndex);

                            //Get the name for the new geodatabase annotation layer.
                            string sNewAnnotationLayerName = AnnotationUtil.GetAnnoLayerName(pAssociatedFeatureLayer);

                            //Check if the annotation layer exists in the database.
                            IWorkspace2 pWorkspace = (IWorkspace2)pFeatureWorkspace;
                            bool bExists = pWorkspace.get_NameExists(esriDatasetType.esriDTFeatureClass, sNewAnnotationLayerName);

                            //************************************************************************
                            //Create the new empty geodatabase annotation layer (and default anno class) 
                            //if it does not already exist.
                            //Otherwise retrieve the existing feature class.
                            //************************************************************************

                            IFeatureClass pFClass = null;
                            if (bExists)
                            {
                                //If the annotation feature class already exists, retrieve it.
                                pFClass = pFeatureWorkspace.OpenFeatureClass(sNewAnnotationLayerName);
                            }
                            else
                            {
                                // Create an annotation class and provide it with a name.
                                ILabelEngineLayerProperties labelEngineLayerProperties = new LabelEngineLayerPropertiesClass();
                                IAnnotateLayerProperties annotateLayerProperties = (IAnnotateLayerProperties)labelEngineLayerProperties;
                                annotateLayerProperties.Class = "Default";

                                //Get the text symbol from the first text element in the old map annotation layer.
                                ITextSymbol annotationTextSymbol;
                                IGraphicsContainer pGraphCont = (IGraphicsContainer)pMapAnnoLayer;
                                pGraphCont.Reset();
                                IElement pOldElement = pGraphCont.Next();
                                if (pOldElement is ITextElement)
                                {
                                    ITextElement pOldTextElement = (ITextElement)pOldElement;
                                    ITextSymbol pOldTextSymbol = pOldTextElement.Symbol;
                                    IClone pTextSymbolClone = (IClone)pOldTextSymbol;
                                    annotationTextSymbol = (ITextSymbol)pTextSymbolClone.Clone();
                                    //Put the new new symbol in the annotation class
                                    labelEngineLayerProperties.Symbol = annotationTextSymbol;
                                }
                                else
                                {
                                    //If the first graphic element is not a text element, we cannot copy its symbology.
                                    //Get the defaul text symbol from the annotation class.
                                    annotationTextSymbol = labelEngineLayerProperties.Symbol;
                                }
                                ISymbol annotationSymbol = (ISymbol)annotationTextSymbol;

                                // Create a symbol collection and add the symbol from the
                                // annotation class to the collection. Assign the resulting symbol ID
                                // to the annotation class.
                                ISymbolCollection symbolCollection = new SymbolCollectionClass();
                                ISymbolCollection2 symbolCollection2 = (ISymbolCollection2)symbolCollection;
                                ISymbolIdentifier2 symbolIdentifier2 = null;
                                symbolCollection2.AddSymbol(annotationSymbol, "Default", out symbolIdentifier2);
                                labelEngineLayerProperties.SymbolID = symbolIdentifier2.ID;

                                // Add the annotation class to a collection.
                                IAnnotateLayerPropertiesCollection annotateLayerPropsCollection = new
                                    AnnotateLayerPropertiesCollectionClass();
                                annotateLayerPropsCollection.Add(annotateLayerProperties);

                                // Create a graphics layer scale object.
                                IGraphicsLayerScale oldGraphicsLayerScale = (IGraphicsLayerScale)pMapAnnoLayer;
                                IGraphicsLayerScale graphicsLayerScale = new GraphicsLayerScaleClass();
                                graphicsLayerScale.ReferenceScale = oldGraphicsLayerScale.ReferenceScale;
                                graphicsLayerScale.Units = pMap.MapUnits;

                                // Create the overposter properties for the standard label engine.
                                IOverposterProperties overposterProperties = new BasicOverposterPropertiesClass();

                                // Instantiate a class description object.
                                IObjectClassDescription ocDescription = new AnnotationFeatureClassDescriptionClass();
                                IFeatureClassDescription fcDescription = (IFeatureClassDescription)ocDescription;

                                // Get the shape field from the class description's required fields.
                                IFields requiredFields = ocDescription.RequiredFields;
                                int shapeFieldIndex = requiredFields.FindField(fcDescription.ShapeFieldName);
                                IField shapeField = requiredFields.get_Field(shapeFieldIndex);
                                IGeometryDef geometryDef = shapeField.GeometryDef;
                                IGeometryDefEdit geometryDefEdit = (IGeometryDefEdit)geometryDef;
                                geometryDefEdit.SpatialReference_2 = pSR;

                                // Create the annotation feature class and an annotation layer for it.
                                IAnnotationLayer annotationLayer = annotationLayerFactory.CreateAnnotationLayer
                                    (pFeatureWorkspace, null, sNewAnnotationLayerName, geometryDef, null,
                                    annotateLayerPropsCollection, graphicsLayerScale, symbolCollection, false, false,
                                    false, true, overposterProperties, "DEFAULTS");

                                // Get the feature class from the feature layer.
                                IFeatureLayer featureLayer = (IFeatureLayer)annotationLayer;
                                pFClass = featureLayer.FeatureClass;

                                //Add a page name field (if it does not already exist)            
                                AnnotationUtil.CreatePageNameField(pFClass);

                                //************************************************************************                        
                                //Finished creating the empty annotation feature class.
                                //************************************************************************
                            }

                            //Determine if this page already has existing annotation features. If not, create them.                            
                            bool bHasAnnoFeaturesForPage = AnnotationUtil.HasAnnoFeaturesForPage(pAssociatedFeatureLayer, pFeatureWorkspace, iPageIndex);
                            if (bHasAnnoFeaturesForPage)
                            {
                                throw new ExceptionUserError("I cannot migrate one or more annotation layers from the map to the geodatabase because "
                                + "the geodatabase already contains annotation features for the corresponding layer and page.");
                            }
                            else
                            {
                                //Open a feature cursor for inserting.
                                IFeatureCursor pFCursor = pFClass.Insert(true);
                                //Loop through each graphic element in the map annotation group,
                                //and copy it to the geodatabase annotation group.
                                //IGraphicsContainer pGC_New = (IGraphicsContainer)annotationLayer;
                                IGraphicsContainer pGC_Old = (IGraphicsContainer)pMapAnnoLayer;
                                pGC_Old.Reset();
                                int iElementCount = 0;
                                IElement pElement;
                                while ((pElement = pGC_Old.Next()) != null)
                                {
                                    iElementCount++;

                                    //Clone the graphic element
                                    IClone pElementClone = ((IClone)pElement).Clone();
                                    IElement pNewElement = (IElement)pElementClone;

                                    //Set up a feature buffer to insert new features
                                    IFeatureBuffer pFeatBuffer = pFClass.CreateFeatureBuffer();
                                    IFeature pFeature = (IFeature)pFeatBuffer;

                                    //Set the AnnotationFeature.Annotation property to newly cloned element
                                    IAnnotationFeature pAnnotationFeature = (IAnnotationFeature)pFeature;
                                    pAnnotationFeature.Annotation = pNewElement;
                                    //Insert the feature
                                    pFCursor.InsertFeature((IFeatureBuffer)pAnnotationFeature);
                                }
                                //Persist the new features
                                pFCursor.Flush();

                                //Create page name attribute if it does not already exists.
                                //Replace all null page name value with the current page name.
                                AnnotationUtil.PopulatePageNameField(pAssociatedFeatureLayer, pFeatureWorkspace, sPageName);
                                
                                //Mark a map annotatio group for deletion.
                                MapAnnoLayersToDelete.Add(sMapAnnoLayerName);
                            }

                            //********************************************************************************
                            //Make a new annotation layer in the TOC (if it does not already exist)
                            //********************************************************************************

                            //If the annotation is already in the TOC, skip this step.
                            IGraphicsLayer pTempLayer = null;
                            if (!AnnotationUtil.HasAnnotationInTOC(pAssociatedFeatureLayer, out pTempLayer))
                            {
                                //Get the annotation feature class
                                IFeatureClass pClass = pFeatureWorkspace.OpenFeatureClass(sNewAnnotationLayerName);
                                IFeatureLayer pAnnoFeatLayer = (IFeatureLayer)new FDOGraphicsLayer();
                                pAnnoFeatLayer.FeatureClass = pClass;
                                pAnnoFeatLayer.Name = pClass.AliasName;

                                //Associate the annotation layer to its feature layer
                                IGraphicsLayer pGraphicsLayer = (IGraphicsLayer)pAnnoFeatLayer;
                                pGraphicsLayer.AssociatedLayer = pAssociatedFeatureLayer;

                                //Generate a new annotation layer
                                //with a page definition query based on a layer file template,
                                //but with other properties that come from the original layer.
                                ILayer pNewAnnoLayer = AnnotationUtil.ModifyAnnoLayerPropertiesFromLyrTemplate(pAnnoFeatLayer);

                                //Add the new annotation layer to the map
                                pMap.AddLayer(pNewAnnoLayer);
                            }

                            //MessageBox.Show("Map Annotation Layer Name: " + pMapAnnoLayer.Name
                            //    + Environment.NewLine + "Element count: " + iElementCount
                            //    + Environment.NewLine + "Associated Feature Layer Name: " + pAssociatedFeatureLayer.Name
                            //    + Environment.NewLine + "New Geodatabase Annotation Layer Name: " + sNewAnnotationLayerName, "JambWow");
                        }
                    }
                }

                //Loop through a list of map annotation layers that were marked for deletion, and delete them.
                ICompositeGraphicsLayer2 pCompositeGraphicsLayer = (ICompositeGraphicsLayer2)pCompositeLayer;
                foreach (string sMapAnnoLayerToDelete in MapAnnoLayersToDelete)
                {
                    pCompositeGraphicsLayer.DeleteLayer(sMapAnnoLayerToDelete);
                }

                //Navigate to the current page.
                int iCurrentPage = DDP_Util.GetCurrentKnownPage();
                DDP_Util.GoToPage(iCurrentPage);

                //Refresh the table of contents
                GenUtil.GetMxDocument().UpdateContents();
            }
            catch (Exception pException)
            {
                if (pException.Message == ANNO_CONVERSION_CANCELED_MESSAGE)
                    MessageBox.Show("Annotation Conversion Canceled. " + Environment.NewLine + " Warning: Partially generated annotation must be deleted manually.", "JambWow - Annotation Conversion Canceled");
                else
                    GenUtil.ShowError(pException);
            }
            finally
            {
                pProgressWindow.Close(); //Close the progress Window.                
            }
        }

        //This procedure clears the selection of the index layer (and all other layers) in the inset map(s).   
        public static void ClearSelectionInInsetMap()
        {
            IMxDocument pMxDoc = GenUtil.GetMxDocument();
            IMaps pMaps = pMxDoc.Maps;
            int MapCount = pMaps.Count;
            if (MapCount > 1)
            {
                for (int x = 1; x < MapCount; x++)
                {
                    IMap pMap = pMaps.Item[x];
                    pMap.ClearSelection();
                }
            }
        }

        //This procedure upgrades old tagged JAMB layout elements.
        //Tag types include:"Show On This Page Only", "Show on All Pages Except This One",
        //"Tag as Data Graph", "Date Tag", "Page Number Tag", "Page Name Tag",
        //and "Extra Item Tag".
        //
        //If a layout element is currently hidde, it cannot be restored to its original position.
        //Instead it will be unhidden and placed below the page.
        //Instead of being re-tagged, old tagging information will be placed into the layout element name.
        //The user must reposition and re-tag these layout elements manually.    
        public static void UpgradeLayoutElements()
        {
            //Clear the selection of the index layer (and all other layers) in the inset map(s).   
            ClearSelectionInInsetMap();
           
            //Various variables used to maintain counts:
            int iNumHiddenElementsMovedBelowPage = 0;
            int iNumHiddenElementsReTaggedThisPageOnly = 0;
            int iNumHiddenElementsReTaggedAllPagesButThis = 0;
            int iDeletedElementCount;
            int iNumTaggedDataGraphs = 0;
            int iNumTaggedDates = 0;
            int iNumTaggedTitles = 0;
            int iNumTaggedPageNumbers = 0;
            int iNumTaggedExtraItems = 0;

            //Clear the selection in the layout view
            IGraphicsContainerSelect pGraphicsContainerSelect = (IGraphicsContainerSelect)ArcMap.Document.PageLayout;
            pGraphicsContainerSelect.UnselectAllElements();

            //Get the page size
            IPageLayout pPageLayout = ArcMap.Document.PageLayout;
            IPage pPage = pPageLayout.Page;
            double dPageHeight = 0;
            double dPageWidth = 0;
            pPage.QuerySize(out dPageWidth, out dPageHeight);

            //Get a clone of the page envelope.
            IClone pEnvelopeClone = (IClone)pPage.PrintableBounds;
            IEnvelope pEnvelope = (IEnvelope)pEnvelopeClone.Clone();

            //Make a list of elements to mark for deletion
            List<IElement> pElementsToDelete = new List<IElement>();

            //Loop through each grapic element in the page layout.
            IGraphicsContainer pGraphCont = (IGraphicsContainer)ArcMap.Document.PageLayout;
            pGraphCont.Reset();
            IElement pElement;
            while ((pElement = pGraphCont.Next()) != null)
            {
                IElementProperties pElementProperties = (IElementProperties)pElement;

                //Mark the element for deletion if it is hidden but has lost its tag.
                if (pElementProperties.Name == ""
                    && pElement.Geometry.Envelope.XMin == HIDE_LOCATION_X
                    && pElement.Geometry.Envelope.YMin == HIDE_LOCATION_Y)
                {
                    pElementsToDelete.Add(pElement);
                }

                //Handle hidden items that are tagged to show on one page or hide on one page.
                if (pElementProperties.Name == TAG_CONST_THISPAGEONLY
                    || pElementProperties.Name == TAG_CONST_ALLPAGESBUTTHIS)
                {
                    int pageNumber = Convert.ToInt32(pElementProperties.Type);
                    if (pElementProperties.CustomProperty is bool)
                    {
                        string sPageNumber = pElementProperties.Type;
                        //If the element is hidden, then move it to a new location below the page.
                        //Its new bottom left corner should be one page height below the bottom left corner of the page.
                        bool bHidden = (bool)pElementProperties.CustomProperty;
                        if (bHidden)
                        {
                            IGeometry pGeometry = pElement.Geometry;
                            ITransform2D pTrans2d = (ITransform2D)pElement.Geometry;
                            pTrans2d.Move((HIDE_LOCATION_X * -1), ((HIDE_LOCATION_Y * -1) - dPageHeight));
                            pElement.Geometry = (IGeometry)pTrans2d;

                            //Increment the count of hidden elements that could not be properly processed.
                            iNumHiddenElementsMovedBelowPage++;

                            //Select the element.                            
                            pGraphicsContainerSelect.SelectElement(pElement);
                                                        
                            //Union the page envelope with the envelope of the hidden element.
                            IClone pElementEnvelopeClone = (IClone)pElement.Geometry.Envelope;
                            IEnvelope pElementEnvelope = (IEnvelope)pElementEnvelopeClone.Clone();
                            pEnvelope.Union(pElementEnvelope);

                            //Re-assign the "Name", "Type", and "CustomProperty" properties.                                                    
                            if (pElementProperties.Name == TAG_CONST_THISPAGEONLY)
                                pElementProperties.Name = PREFIX_THISPAGEONLY + sPageNumber;
                            if (pElementProperties.Name == TAG_CONST_ALLPAGESBUTTHIS)
                                pElementProperties.Name = PREFIX_ALLPAGESBUTTHIS + sPageNumber;
                            pElementProperties.Type = "";
                            pElementProperties.CustomProperty = null;
                        }
                        else
                        {
                            
                            //Get the old tag name.
                            string sOldTag = pElementProperties.Name;

                            //Clear the "Name", "Type", and "CustomProperty" properties.                                                    
                            pElementProperties.Name = "";
                            pElementProperties.Type = "";
                            pElementProperties.CustomProperty = null;

                            //Re-tag the layout element (using the JambWow method) to be hidden on a subset of pages.
                            if (sOldTag == TAG_CONST_THISPAGEONLY)
                            {
                                LayoutElementTaggingUtil.TagToShowOnThisPageOnly(pElementProperties, pageNumber);
                                //Increment the count of elements that were correctly re-tagged as "Show On This Page Only"
                                iNumHiddenElementsReTaggedThisPageOnly++;
                            }
                            else if (sOldTag == TAG_CONST_ALLPAGESBUTTHIS)
                            {
                                LayoutElementTaggingUtil.TagToShowOnAllPagesButThis(pElementProperties, pageNumber);
                                //Increment the count of elements that were correctly re-tagged as "Show On All Pages But This One"
                                iNumHiddenElementsReTaggedAllPagesButThis++;
                            }
                        }
                    }
                }

                //Handle re-tagging of data graph(s).
                if (pElementProperties.Type == TAG_CONST_DATAGRAPH)
                {
                    iNumTaggedDataGraphs++;
                    //Clear the "Type" property.      
                    pElementProperties.Type = "";                                               
                    //Tag the layout element as a data graph
                    LayoutElementTaggingUtil.TagOrUntagDataGraph(true, pElementProperties);                    
                }
                                
                //Handle re-tagging of Date (as dynamic text)        
                if (pElementProperties.Name == TAG_CONST_DATE || pElementProperties.Name == TAG_CONST_DATE2)
                {
                    iNumTaggedDates++;
                    //Clear the name property
                    pElementProperties.Name = "";
                    //Set the text to display the date using dynamic text.
                    if (pElementProperties is ITextElement)
                    {
                        ITextElement pTextElement = (ITextElement) pElementProperties;
                        pTextElement.Text = @"<dyn type=""date"" format=""short""/>";
                    }
                }

                //Handle re-tagging of Page Name/Title (as dynamic text)
                if (pElementProperties.Name == TAG_CONST_TITLE)
                {
                    iNumTaggedTitles++;
                    //Clear the name property
                    pElementProperties.Name = "";
                    //Set the text to display the page name/title using dynamic text.
                    if (pElementProperties is ITextElement)
                    {
                        ITextElement pTextElement = (ITextElement)pElementProperties;
                        pTextElement.Text = @"<dyn type=""page"" property=""name""/>";
                    }

                }

                //Handle re-tagging of Page Number (as dynamic text)
                if (pElementProperties.Name == TAG_CONST_PAGENUMBER)
                {
                    iNumTaggedPageNumbers++;
                    //Clear the name property
                    pElementProperties.Name = "";

                    //Set the text to display the page number using dynamic text.
                    if (pElementProperties is ITextElement)
                    {
                        ITextElement pTextElement = (ITextElement)pElementProperties;
                        pTextElement.Text = @"<dyn type=""page"" property=""number""/>";
                    }
                }

                //Handle re-tagging of Extra Items
                if (pElementProperties.Name == TAG_CONST_EXTRAITEM)
                {
                    iNumTaggedExtraItems++;
                    string sAttributeName = pElementProperties.Type;
                    //Clear the name and type properties
                    pElementProperties.Name = "";
                    pElementProperties.Type = "";

                    //Set the text to display the page number using dynamic text.
                    if (pElementProperties is ITextElement)
                    {
                        ITextElement pTextElement = (ITextElement)pElementProperties;
                        pTextElement.Text = @"<dyn type=""page"" property=""" + sAttributeName + @"""/>";
                    }
                }
            }

            //Delete any elements that were marked for deletion
            iDeletedElementCount = pElementsToDelete.Count;
            foreach (IElement pElementToDelete in pElementsToDelete)
                pGraphCont.DeleteElement(pElementToDelete);

            //If hidden elements were found and moved below the page, zoom to the combined envelope of the hidden elements and the page.
            if (iNumHiddenElementsMovedBelowPage > 0)
            {
                IActiveView pActiveLayoutView = (IActiveView)ArcMap.Document.PageLayout;
                pActiveLayoutView.Extent = pEnvelope;
            }

            //Refresh the page.
            ((IActiveView)ArcMap.Document.PageLayout).Refresh(); 
            
            //********************************************************************
            //Display a migration report.            
            //Modify the code here to display the report in an ASCII file.
            //Maybe only show items that have a count of 1 or more.
            //********************************************************************
            string sMigrationReport = "Migration Report:" + Environment.NewLine;
            if (iNumHiddenElementsMovedBelowPage > 0)
            {
                sMigrationReport = sMigrationReport + Environment.NewLine
                    + "Action Item!:  " + iNumHiddenElementsMovedBelowPage + "  hidden layout element(s) were placed below the page because thier location information was lost."
                    + Environment.NewLine + "Please manually move these layout elements to the correct position and then re-tag."
                    + Environment.NewLine + "The previously tagged page numbers can be viewed in the Element Name box in the Properties window under"
                    + Environment.NewLine + "the Size and Position tab." + Environment.NewLine;
            }
            if (iNumTaggedDataGraphs > 0)
                sMigrationReport = sMigrationReport + Environment.NewLine + iNumTaggedDataGraphs + "  Tagged Data Graph(s) Migrated";
            if (iNumTaggedTitles > 0)
                sMigrationReport = sMigrationReport + Environment.NewLine + iNumTaggedTitles + "  Tagged Page Title(s) Migrated";
            if (iNumTaggedPageNumbers > 0)
                sMigrationReport = sMigrationReport + Environment.NewLine + iNumTaggedPageNumbers + "  Tagged Page Number(s) Migrated";
            if (iNumTaggedExtraItems > 0)
                sMigrationReport = sMigrationReport + Environment.NewLine + iNumTaggedExtraItems + "  Tagged Index Layer Attribute(s) Migrated";
            if (iNumTaggedDates > 0)
                sMigrationReport = sMigrationReport + Environment.NewLine + iNumTaggedDates + "  Tagged Date Element(s) Migrated";
            if (iNumHiddenElementsReTaggedThisPageOnly > 0)
                sMigrationReport = sMigrationReport + Environment.NewLine + iNumHiddenElementsReTaggedThisPageOnly + "  'Show On This Page Only' Item(s) Migrated";
            if (iNumHiddenElementsReTaggedAllPagesButThis > 0)
                sMigrationReport = sMigrationReport + Environment.NewLine + iNumHiddenElementsReTaggedAllPagesButThis + "  'Show On All Pages But This One' Item(s) Migrated";
            if (iDeletedElementCount > 0)
            {
                sMigrationReport = sMigrationReport + Environment.NewLine + Environment.NewLine
                    + "Note:  " + iDeletedElementCount + "  hidden layout element(s) were deleted because of broken tags.";
            }

            //Create a temp text file to store the report
            string sReportFileName = JAMBWOW_Extension.TempFiles.CreateTempFile("JambWow_Migration_Report", "txt");
            GenUtil.WriteToFile(sReportFileName, sMigrationReport);

            //Show a message box
            MessageBox.Show("Finished Migrating Old JAMB Tagged Elements. A migration report will be displayed in your default text editor.", "JambWow");

            //Open the text file using the default text file viewer
            Process myProc;
            myProc = Process.Start(sReportFileName);
        }
    }
}
