﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using ESRI.ArcGIS.Carto;
using ESRI.ArcGIS.ArcMapUI;
using ESRI.ArcGIS.esriSystem;
using ESRI.ArcGIS.Geometry;
using ESRI.ArcGIS.Display;

namespace JAMBWOW
{
    public static class LayoutElementTaggingUtil
    {        
        private const string TAG_NAME_DATAGRAPH = "DataGraph";
        
        private const string TAG_HIDDEN_ON_PAGES = "HiddenOnPages";
        private const string TAG_CURRENTLY_HIDDEN = "CurrentlyHidden";

        private const string TAG_PAGE_SPECIFIC_GEOMETRIES = "PageSpecificGeometries";
        private const string TAG_STANDARD_GEOMETRY = "StandardGeometry";
        private const string TAG_STANDARD_EXTRA = "StandardExtra";
        private const string TAG_ALT_EXTRA = "Extra";
        private const string TAG_TEXT_SIZE = "TextSize";
        private const string TAG_CALLOUT_ANCHOR_POINT = "CalloutAnchorPoint";
        private const string TAG_NORTH_ARROW_SIZE = "NorthArrowSize";

        private const int HIDE_LOCATION_X_OFFSET = -100;
        private const int HIDE_LOCATION_Y_OFFSET = 0;

        #region Public Methods

        //This method updates tagged layout elements.
        //It is called before navigating to a page.
        //It is sometimes called after layout elements are tagged/untagged.
        public static void UpdateTaggedElements(int pageIndex, bool bPartialRefresh)
        {
            try
            {
                IMxDocument pDoc = GenUtil.GetMxDocument();
                IActiveView pLayout = (IActiveView)pDoc.PageLayout;
                IGraphicsContainer pGraphicsCont = (IGraphicsContainer)pDoc.PageLayout;
                pGraphicsCont.Reset();

                //Get the current page name
                string sCurrentPageName = IndexLayerUtil.GetPageName(pageIndex); //Get the page name.

                //Loop through each of the layout elements
                IElementProperties pElemProps = null;
                while ((pElemProps = (IElementProperties)pGraphicsCont.Next()) != null)
                {                                        
                    if (pElemProps.CustomProperty is IPropertySet2) //If the layout element contains tag
                    {
                        IElement pElement = (IElement)pElemProps;

                        //Get the envelope of the layout element before it is modified.
                        IEnvelope pEnv = new EnvelopeClass();
                        pElement.QueryBounds(pLayout.ScreenDisplay, pEnv);

                        IPropertySet2 pPropertySet = (IPropertySet2)pElemProps.CustomProperty;
                        bool bUpdate = false; //Used to determine if we should update the display of the layout element.

                        //Determine if this is a tagged data graph
                        if (PropertySetUtil.ContainsProperty(pPropertySet, TAG_NAME_DATAGRAPH))
                        {
                            DataGraphUtil.SetPropertiesOfDataGraph(pElemProps, pageIndex);
                            bUpdate = true;
                        }

                        //Determine if this layout element is tagged with standard geometry information,
                        //and if this page contains alternate geometry information.
                        IPropertySet2 PageSpecificGeometriesPropertySet = PropertySetUtil.GetInnerPropSet(pPropertySet, TAG_PAGE_SPECIFIC_GEOMETRIES);
                        IGeometry pStandardGeometry = (IGeometry) PropertySetUtil.GetPropertyValue<IGeometry>(
                            PageSpecificGeometriesPropertySet, TAG_STANDARD_GEOMETRY);
                        
                        if (pStandardGeometry != null)
                        {
                            MoveAndResizeLayoutElement(ref pElement, PageSpecificGeometriesPropertySet, sCurrentPageName, pStandardGeometry);
                            //After loading a stored geometry, mark the layout element as not hidden.
                            pPropertySet.RemoveProperty(TAG_CURRENTLY_HIDDEN);
                            bUpdate = true; //Update the display.
                        }
                        
                        //Determine if this layout element is tagged as currently hidden
                        bool bCurrentlyHidden = PropertySetUtil.ContainsProperty(pPropertySet, TAG_CURRENTLY_HIDDEN);
                        
                        //Determine if this layout element should be shown or hidden on the current page.                        
                        IPropertySet2 PS_HiddenOnPagesT = PropertySetUtil.GetInnerPropSet(pPropertySet, TAG_HIDDEN_ON_PAGES);
                        bool bShowOnCurrentPage = !PropertySetUtil.ContainsProperty(PS_HiddenOnPagesT, sCurrentPageName);
                                                
                        //Hide or show this layout element as needed.                                                
                        if (bCurrentlyHidden && bShowOnCurrentPage)
                        {
                            ShowLayoutElement(ref pElement); //Show the layout element
                            pPropertySet.RemoveProperty(TAG_CURRENTLY_HIDDEN); //Remove the "currently hidden" tag.
                            bUpdate = true; //Update the display.
                        }
                        else if (!bCurrentlyHidden && !bShowOnCurrentPage)
                        {
                            HideLayoutElement(ref pElement); //Hide the layout element
                            pPropertySet.SetProperty(TAG_CURRENTLY_HIDDEN, TAG_CURRENTLY_HIDDEN); //Set the "currently hidden" tag.
                            bUpdate = true; //Update the display.
                        }

                        //If the geometry of the tagged element has changed,
                        //and "Partial Refresh" is set to true, refresh the element in the layout.
                        if (bUpdate && bPartialRefresh)
                        {
                            //Update the element
                            pGraphicsCont.UpdateElement(pElement);                            
                            //Get the envelope of the layout element before after it is modified,
                            //and unition it with the old envelope.
                            IEnvelope pEnv2 = new EnvelopeClass();
                            pElement.QueryBounds(pLayout.ScreenDisplay, pEnv2);
                            pEnv.Union(pEnv2);
                            //Refresh the graphics and graphic selection within the unioned envelope
                            pLayout.PartialRefresh(esriViewDrawPhase.esriViewGraphics, null, pEnv);
                            pLayout.PartialRefresh(esriViewDrawPhase.esriViewGraphicSelection, null, pEnv);
                        }
                    }
                }
            }
            catch (Exception e)
            {
                GenUtil.ShowError(e);
            }
        }

        //This method is called to tag or untag a datagraph
        public static void TagOrUntagDataGraph(bool bTag, IElementProperties pElemProps = null)
        {
            try
            {
                //Get the selected layout element (if it was not passed in)
                if (pElemProps == null) pElemProps = GetSelectedLayoutElement();
                //Exit if the selected layout element is not a datagraph
                if (!(pElemProps is IDataGraphTElement))
                {
                    MessageBox.Show("The selected element is not a data graph.", "JambWow - Could not tag layout element.");
                    return;
                }
                //Retrieve the property set or create a new property set.
                IPropertySet2 pPropertySet = GetPropertySet(pElemProps);
                //Tag or un-tag the data graph
                if (bTag)
                {
                    pPropertySet.SetProperty(TAG_NAME_DATAGRAPH, (object)TAG_NAME_DATAGRAPH); //Create the datagraph tag
                    DDP_Util.PageChangeActions(DDP_Util.GetCurrentKnownPage(), true); //Run page change actions.
                }
                else
                    pPropertySet.RemoveProperty(TAG_NAME_DATAGRAPH);  //Remove the datagraph tag
                //Update layout element's custom property
                pElemProps.CustomProperty = (object)pPropertySet;
            }
            catch (Exception e)
            {
                GenUtil.ShowError(e);
            }
        }

        //This method is called to untag a layout element to show on all pages.
        public static void UntagToShowOnAllPages()
        {
            try
            {
                //Get the selected layout element (if an element is not passed in).
                IElementProperties pElemProps = GetSelectedLayoutElement();                
                //Exit if there is not a single selected layout element.
                if (pElemProps == null) return;
                //Retrieve the property set or create a new property set.
                IPropertySet2 pPropertySet = GetPropertySet(pElemProps);
                //Update the top-level property set with an updated list of pages to hide.                
                pPropertySet.RemoveProperty(LayoutElementTaggingUtil.TAG_HIDDEN_ON_PAGES);
                //Set the custom property of the layout element to the updated property set.
                pElemProps.CustomProperty = (object)pPropertySet;
            }
            catch (Exception e)
            {
                GenUtil.ShowError(e);
            }
        }

        //This method is called to tag a layout element to be hidden on the current page only.
        public static void TagToShowOnAllPagesButThis(IElementProperties pElemProps = null, int? pageIndex = null)
        {
            try
            {
                //Get the current page index (if a page index was not passed if).
                if (pageIndex == null) pageIndex = DDP_Util.GetCurrentKnownPage();

                //Get the selected layout element (if an element is not passed in).
                if (pElemProps == null) pElemProps = GetSelectedLayoutElement();
                //Exit if there is not a single selected layout element.
                if (pElemProps == null) return;
                //Retrieve the property set or create a new property set.
                IPropertySet2 pPropertySet = GetPropertySet(pElemProps);

                //Get the name of the page
                string sPageName = IndexLayerUtil.GetPageName((int)pageIndex);

                //Create a new propertyset to hold the list of pages to be hidden on.
                IPropertySet2 NamesOfPagesHiddenOn = new PropertySetClass();
                //Add the page name to the propertyset.
                NamesOfPagesHiddenOn.SetProperty(sPageName, (object)sPageName);
                //Update the top-level property set with an updated list of pages to hide.                
                pPropertySet.SetProperty(LayoutElementTaggingUtil.TAG_HIDDEN_ON_PAGES, NamesOfPagesHiddenOn);
                //Set the custom property of the layout element to the updated property set.
                pElemProps.CustomProperty = (object)pPropertySet;
            }
            catch (Exception e)
            {
                GenUtil.ShowError(e);
            }
        }

        //This method is called to tag a layout element to be hidden on all pages except the current page.
        public static void TagToShowOnThisPageOnly(IElementProperties pElemProps = null, int? pageIndex = null)
        {
            try
            {
                //Get the current page index (if a page index was not passed if).
                if (pageIndex == null) pageIndex = DDP_Util.GetCurrentKnownPage();

                //Get the selected layout element (if an element is not passed in).
                if (pElemProps == null) pElemProps = GetSelectedLayoutElement();
                //Exit if there is not a single selected layout element.
                if (pElemProps == null) return;
                //Retrieve the property set or create a new property set.
                IPropertySet2 pPropertySet = GetPropertySet(pElemProps);

                //Get an array of page names from IndexLayerUtil.
                string[] ArrPageNames = IndexLayerUtil.GetPageNames(); //Get the names of the pages                
                //Validate that the number of pages returned matches with Data Driven Pages.
                int pageCount = DDP_Util.GetPageCount(); //Get page count.                
                if (ArrPageNames == null) return;
                if (ArrPageNames.GetLength(0) == 0) return;
                if (ArrPageNames.GetLength(0) != pageCount)
                {
                    MessageBox.Show("The number of pages in your index layer does not match the number of pages in Data Driven Pages. Refresh DDP and try again.", "JambWow");
                    return;
                }

                //Get the name of the current page
                string sCurrentPageName = IndexLayerUtil.GetPageName((int)pageIndex);
                //Create a new propertyset to hold the list of pages to be hidden on.
                IPropertySet2 NamesOfPagesHiddenOn = new PropertySetClass();
                //Loop through the array of page names. Add all of the page names to the propertset except the current page.
                for (int n = 1; n <= pageCount; n++)
                {
                    string PageName = ArrPageNames[n - 1];
                    if (PageName != sCurrentPageName)
                        NamesOfPagesHiddenOn.SetProperty(PageName, (object)PageName);
                }
                //Update the top-level property set with an updated list of pages to hide.                
                pPropertySet.SetProperty(LayoutElementTaggingUtil.TAG_HIDDEN_ON_PAGES, NamesOfPagesHiddenOn);
                //Set the custom property of the layout element to the updated property set.
                pElemProps.CustomProperty = (object)pPropertySet;
            }
            catch (Exception e)
            {
                GenUtil.ShowError(e);
            }
        }

        //This method is called to tag a layout element to be hidden on a subset of pages.
        public static void TagToHideOnSelectedPages()
        {
            try
            {
                //Get the selected layout element (if an element is not passed in).
                IElementProperties pElemProps = GetSelectedLayoutElement();
                //Exit if there is not a single selected layout element.
                if (pElemProps == null) return;
                //Retrieve the property set or create a new property set.
                IPropertySet2 pPropertySet = GetPropertySet(pElemProps);

                //Get an array of page names from IndexLayerUtil.
                string[] ArrPageNames = IndexLayerUtil.GetPageNames(); //Get the names of the pages                
                //Validate that the number of pages returned matches with Data Driven Pages.
                int pageCount = DDP_Util.GetPageCount(); //Get page count.                
                if (ArrPageNames == null) return;
                if (ArrPageNames.GetLength(0) == 0) return;
                if (ArrPageNames.GetLength(0) != pageCount)
                {
                    MessageBox.Show("The number of pages in your index layer does not match the number of pages in Data Driven Pages. Refresh DDP and try again.", "JambWow");
                    return;
                }

                //Get an array of page numbers
                string[] ArrPageNumbers = IndexLayerUtil.GetPageNumbers();
                if (ArrPageNumbers == null) return;
                if (ArrPageNumbers.GetLength(0) == 0) return;
                if (ArrPageNumbers.GetLength(0) != pageCount)
                {
                    MessageBox.Show("The number of pages in your index layer does not match the number of pages in Data Driven Pages. Refresh DDP and try again.", "JambWow");
                    return;
                }

                //Build a list of Page Wrapper objects that contain both page name and page index.
                List<PageWrapper> PageWrappers = new List<PageWrapper>();
                for (int n = 1; n <= pageCount; n++)
                {
                    string PageName = ArrPageNames[n - 1];
                    string sPageNumber = ArrPageNumbers[n - 1];
                    PageWrapper pPageWrapper = new PageWrapper(PageName, n, sPageNumber);
                    PageWrappers.Add(pPageWrapper);
                }

                //Determine if this layout element contains an existing tag to hide it on some pages. If so, get the page names.                    
                IPropertySet2 NamesOfPagesHiddenOn = PropertySetUtil.GetInnerPropSet(pPropertySet, TAG_HIDDEN_ON_PAGES);
                
                //Convert PropertySet to a list of strings               
                object names = null, values = null;
                NamesOfPagesHiddenOn.GetAllProperties(out names, out values);
                List<string> Str_List_NamesOfPagesHiddenOn = new List<string>();
                string[] nameArray = (string[])names;
                for (int i = 0; i < nameArray.GetLength(0); ++i)
                    Str_List_NamesOfPagesHiddenOn.Add(nameArray[i]);

                //Build one list of Page Wrapper objects for pages on which the layout element is hidden
                //and build another list of Page Wrapper objects for pages on which the layout element is not hidden.
                List<PageWrapper> HiddenOnPages = new List<PageWrapper>();
                List<PageWrapper> NotHiddenOnPages = new List<PageWrapper>();
                foreach (PageWrapper pPageWrapper in PageWrappers)
                {
                    if (Str_List_NamesOfPagesHiddenOn.Contains(pPageWrapper.PageName))
                        HiddenOnPages.Add(pPageWrapper);
                    else
                        NotHiddenOnPages.Add(pPageWrapper);
                }

                //Show the "Selected Pages" Form.
                FormSelectedPages pFormSelectedPages = new FormSelectedPages(HiddenOnPages, NotHiddenOnPages);
                pFormSelectedPages.ShowDialog(GenUtil.GetAppWinRef());

                //If the user closed the dialog by clicking "OK", update the propertyset with a new list of pages to hide on.
                if (pFormSelectedPages.DialogResult == DialogResult.OK)
                {
                    //Get a list of names of pages on which the layout element should be hidden.
                    List<string> NewPagesHiddenOn = pFormSelectedPages.GetPagesHiddenOn();
                    //Place the names in a property set.
                    NamesOfPagesHiddenOn = new PropertySetClass();
                    foreach (string sName in NewPagesHiddenOn)
                    {
                        NamesOfPagesHiddenOn.SetProperty(sName, (object)sName);
                    }
                    //Update the top-level property set with an updated list of pages to hide.
                    pPropertySet.SetProperty(LayoutElementTaggingUtil.TAG_HIDDEN_ON_PAGES, NamesOfPagesHiddenOn);                    
                    //Set the custom property of the layout element to the updated property set.
                    pElemProps.CustomProperty = (object)pPropertySet;
                    
                    //Run page change actions.
                    DDP_Util.PageChangeActions(DDP_Util.GetCurrentKnownPage(), true);
                }                
            }
            catch (Exception e)
            {
                GenUtil.ShowError(e);
            }
        }

        //This method is called to tag a layout element with a standard geometry
        //or to remove geometry tag information from a layout element.
        public static void TagOrUntagStandardGeometry(bool bTag)
        {
            try
            {
                //Get the selected layout element
                IElementProperties pElemProps = GetSelectedLayoutElement();
                //Exit if there is not a single selected layout element.
                if (pElemProps == null) return;
                //Retrieve the property set or create a new property set.
                IPropertySet2 pPropertySet = GetPropertySet(pElemProps);

                //If we are untagging, simply remove the TAG_PAGE_SPECIFIC_GEOMETRIES tag. 
                if (!bTag) pPropertySet.RemoveProperty(TAG_PAGE_SPECIFIC_GEOMETRIES);
                else //Otherwise, store the standard geometry inside a new or existing TAG_PAGE_SPECIFIC_GEOMETRIES tag.
                {
                    //If this layout element contains an existing PAGE_SPECIFIC_GEOMETRIES tag, get it. Otherwise, create a new one.
                    IPropertySet2 PageSpecificGeometriesPropertySet = PropertySetUtil.GetInnerPropSet(pPropertySet, TAG_PAGE_SPECIFIC_GEOMETRIES);
                                           
                    //Clone the geometry of the layout element
                    IElement pElement = (IElement)pElemProps;
                    IGeometry pGeometry = pElement.Geometry;
                    IClone pOriginalGeom = (IClone)pGeometry;
                    IClone pCopyGeom = pOriginalGeom.Clone();
                    IGeometry pGeometryCopy = (IGeometry)pCopyGeom;

                    //Store the standard geometry as a child tag under the PAGE_SPECIFIC_GEOMETRIES tag.                
                    PageSpecificGeometriesPropertySet.SetProperty(TAG_STANDARD_GEOMETRY, (object)pGeometryCopy);

                    //Store additional position and size information that is specific to certain element types.
                    StoreAdditionalSizeAndPositionInfo(pElement, ref PageSpecificGeometriesPropertySet, TAG_STANDARD_EXTRA);

                    //Mark the page layout element as not currently hidden.
                    pPropertySet.RemoveProperty(TAG_CURRENTLY_HIDDEN);

                    //Update the top-level property set 
                    pPropertySet.SetProperty(TAG_PAGE_SPECIFIC_GEOMETRIES, PageSpecificGeometriesPropertySet);
                    //Set the custom property of the layout element to the updated property set.
                    pElemProps.CustomProperty = (object)pPropertySet;
                }
            }
            catch (Exception e)
            {
                GenUtil.ShowError(e);
            }
        }

        //This method is called to tag a layout element with an alternate geometry for a particular page
        //or to restore the geometry for that page to the standard geometry.
        public static void TagOrUntagAlternateGeometry(bool bTag)
        {
            try
            {
                //Get the selected layout element
                IElementProperties pElemProps = GetSelectedLayoutElement();
                //Exit if there is not a single selected layout element.
                if (pElemProps == null) return;
                //Retrieve the property set or create a new property set.
                IPropertySet2 pPropertySet = GetPropertySet(pElemProps);

                //Determine if this layout element contains an existing PAGE_SPECIFIC_GEOMETRIES tag.
                //Also determine if the PAGE_SPECIFIC_GEOMETRIES tag contains a standard geometry.                              
                IPropertySet2 PageSpecificGeometriesPropertySet = PropertySetUtil.GetInnerPropSet(pPropertySet, TAG_PAGE_SPECIFIC_GEOMETRIES);
                IGeometry pStandardGeometry = (IGeometry) PropertySetUtil.GetPropertyValue<IGeometry>(
                    PageSpecificGeometriesPropertySet, TAG_STANDARD_GEOMETRY);
                
                //If the PAGE_SPECIFIC_GEOMETRIES tag does not exists or it does not contain a standard geometry, show an error.
                if (pStandardGeometry == null)
                {
                    MessageBox.Show("The selected layout element is not tagged with a standard position and size."
                        + Environment.NewLine + "I cannot tag it with an alternate position and size for this page.", "JambWow");                    
                    return;
                }

                //Get the name of the current page
                //string sCurrentPageName = DDP_Util.GetCurrentPageName();
                string sCurrentPageName = IndexLayerUtil.GetPageName(DDP_Util.GetCurrentKnownPage());

                //Tag alternate geometry or remove alternate geometry tag
                if (bTag)
                {
                    //Clone the geometry of the layout element
                    IElement pElement = (IElement)pElemProps;
                    IGeometry pGeometry = pElement.Geometry;
                    IClone pOriginalGeom = (IClone)pGeometry;
                    IClone pCopyGeom = pOriginalGeom.Clone();
                    IGeometry pGeometryCopy = (IGeometry)pCopyGeom;
                                        
                    //Store the alternate geometry as a child tag under the PAGE_SPECIFIC_GEOMETRIES tag.
                    PageSpecificGeometriesPropertySet.SetProperty(sCurrentPageName, (object)pGeometryCopy);

                    //Store additional position and size information that is specific to certain element types.
                    StoreAdditionalSizeAndPositionInfo(pElement, ref PageSpecificGeometriesPropertySet, sCurrentPageName + TAG_ALT_EXTRA);
                }
                else
                {
                    //Remove alternate geometry tag
                    PageSpecificGeometriesPropertySet.RemoveProperty(sCurrentPageName);
                    //Remove alternate tag that contains additional position and size information
                    PageSpecificGeometriesPropertySet.RemoveProperty(sCurrentPageName + TAG_ALT_EXTRA);

                    //Restore layout element to the standard position and size
                    IElement pElement = (IElement)pElemProps;
                    //Get the "StandardExtra" Propertyset if it exists
                    IPropertySet2 ExtraPropertySet = (IPropertySet2)PropertySetUtil.GetPropertyValue<IPropertySet2>(
                        PageSpecificGeometriesPropertySet, TAG_STANDARD_EXTRA);
                    UpdateElementSizeAndPosition(ref pElement, pStandardGeometry, ExtraPropertySet);
                }

                //Mark the page layout element as not currently hidden.
                pPropertySet.RemoveProperty(TAG_CURRENTLY_HIDDEN);

                //Update the top-level property set 
                pPropertySet.SetProperty(TAG_PAGE_SPECIFIC_GEOMETRIES, PageSpecificGeometriesPropertySet);
                //Set the custom property of the layout element to the updated property set.
                pElemProps.CustomProperty = (object)pPropertySet;

                //Run page change actions.
                DDP_Util.PageChangeActions(DDP_Util.GetCurrentKnownPage(), true);
            }
            catch (Exception e)
            {
                GenUtil.ShowError(e);
            }
        }

        #endregion

        #region Private Methods

        //Store Additional Size and Position Information for a layout element
        private static void StoreAdditionalSizeAndPositionInfo(IElement pElement,
            ref IPropertySet2 PageSpecificGeometriesPropertySet, string NameOfExtraTag)
        {
            //If the layout element is a text element, store additional information.                    
            if (pElement is ITextElement)
            {
                //Create a child tag/propertyset to store additional information.
                IPropertySet2 ExtraPropertySet = new PropertySetClass();
                PageSpecificGeometriesPropertySet.SetProperty(NameOfExtraTag, ExtraPropertySet);

                //Store the text size.
                ITextElement pTextElement = (ITextElement)pElement;
                ExtraPropertySet.SetProperty(TAG_TEXT_SIZE, pTextElement.Symbol.Size);

                //If the text element has a callout, store the callout anchor point.
                if (pTextElement.Symbol is IFormattedTextSymbol)
                {
                    IFormattedTextSymbol pFormattedTextSymbol = (IFormattedTextSymbol)pTextElement.Symbol;
                    if (pFormattedTextSymbol.Background is ICallout)
                    {
                        ICallout pCallout = (ICallout)pFormattedTextSymbol.Background;
                        IPoint pAnchorPoint = pCallout.AnchorPoint;
                        //Clone the anchor point
                        IClone pOriginalPoint = (IClone)pAnchorPoint;
                        IClone pCopyPoint = pOriginalPoint.Clone();
                        IPoint pAnchorPointCopy = (IPoint)pCopyPoint;
                        //Store a copy of the anchor point.
                        ExtraPropertySet.SetProperty(TAG_CALLOUT_ANCHOR_POINT, pAnchorPointCopy);
                    }
                }
            }

            //If the layout element is a north arrow, store additional information.
            if (pElement is IMapSurroundFrame && ((IMapSurroundFrame)pElement).MapSurround is INorthArrow)
            {
                //Create a child tag/propertyset to store additional information.
                IPropertySet2 ExtraPropertySet = new PropertySetClass();
                PageSpecificGeometriesPropertySet.SetProperty(NameOfExtraTag, ExtraPropertySet);

                //Store the north arrow size.
                INorthArrow pNorthArrow = (INorthArrow)((IMapSurroundFrame)pElement).MapSurround;
                ExtraPropertySet.SetProperty(TAG_NORTH_ARROW_SIZE, pNorthArrow.Size);
            }
        }

        //Move and Resize layout element
        private static void MoveAndResizeLayoutElement(ref IElement pElement, IPropertySet2 PageSpecificGeometriesPropertySet,
            string sCurrentPageName, IGeometry pStandardGeometry)
        {
            //Determine if this page contains alternate geometry information
            IGeometry pAlternateGeometry = (IGeometry)PropertySetUtil.GetPropertyValue<IGeometry>(
                PageSpecificGeometriesPropertySet, sCurrentPageName);

            //Determine if the alternate position and size information exits for this page.
            //If not, use standard position and size information instead.
            //Also load any additiona attributes specific to certain element types.
            IGeometry pGeometry;
            IPropertySet2 ExtraPropertySet;
            if (pAlternateGeometry == null) //Standard Position and size
            {
                pGeometry = pStandardGeometry;
                //Get the "StandardExtra" Propertyset if it exists
                ExtraPropertySet = (IPropertySet2)PropertySetUtil.GetPropertyValue<IPropertySet2>(
                    PageSpecificGeometriesPropertySet, TAG_STANDARD_EXTRA);
            }
            else //Alternate position and size
            {
                pGeometry = pAlternateGeometry;
                //Get the <Page Name>"Extra" Propertyset if it exits
                ExtraPropertySet = (IPropertySet2)PropertySetUtil.GetPropertyValue<IPropertySet2>(
                    PageSpecificGeometriesPropertySet, sCurrentPageName + TAG_ALT_EXTRA);
            }
            UpdateElementSizeAndPosition(ref pElement, pGeometry, ExtraPropertySet);
        }

        //Update an element geometry and additional size and position information
        private static void UpdateElementSizeAndPosition(ref IElement pElement, IGeometry pGeometry, IPropertySet2 ExtraPropertySet)
        {
            //Replace the layout element geometry with the alternate geometry or standard geometry.
            IClone pOriginalGeom = (IClone)pGeometry;
            IClone pCopyGeom = pOriginalGeom.Clone();
            IGeometry pGeometryCopy = (IGeometry)pCopyGeom;
            //Set the element geometry to the clone.
            pElement.Geometry = pGeometryCopy;

            //If additional size and position information was stored, update the element.
            if (ExtraPropertySet != null)
            {
                //If element is a text element
                if (pElement is ITextElement)
                {
                    ITextElement pTextElement = (ITextElement)pElement;

                    //If ExtraPropertySet contains text size, set the text size in the text element
                    object oTextSize = PropertySetUtil.GetPropertyValue<double>(ExtraPropertySet, TAG_TEXT_SIZE);
                    if (oTextSize != null)
                    {
                        double dTextSize = (double)oTextSize;
                        ITextSymbol pTextSymbol2 = pTextElement.Symbol;
                        pTextSymbol2.Size = dTextSize;
                        pTextElement.Symbol = pTextSymbol2;
                    }

                    //If the text element contains a callout anchor point
                    ITextSymbol pTextSymbol = pTextElement.Symbol;
                    if (pTextSymbol is IFormattedTextSymbol)
                    {
                        IFormattedTextSymbol pFormattedTextSymbol = (IFormattedTextSymbol)pTextSymbol;
                        ITextBackground pTextBackGround = pFormattedTextSymbol.Background;
                        if (pTextBackGround is ICallout)
                        {
                            //If ExtraPropertySet contains a callout anchor point, set the callout anchor point in the text element
                            IPoint pPoint = (IPoint) PropertySetUtil.GetPropertyValue<IPoint>(ExtraPropertySet, TAG_CALLOUT_ANCHOR_POINT);
                            if (pPoint != null)
                            {
                                ICallout pCallout = (ICallout)pTextBackGround;
                                //Update the anchor point, text background, text symbol, and text element.
                                pCallout.AnchorPoint = pPoint;
                                pFormattedTextSymbol.Background = (ITextBackground)pCallout;
                                pTextElement.Symbol = (ITextSymbol)pFormattedTextSymbol;
                            }
                        }
                    }
                }

                //If element is a north arrow
                if (pElement is IMapSurroundFrame && ((IMapSurroundFrame)pElement).MapSurround is INorthArrow)
                {
                    //If ExtraPropertySet contains north arrow size, update the north arrow size of the layout element
                    object oNorthArrowSize = PropertySetUtil.GetPropertyValue<double>(ExtraPropertySet, TAG_NORTH_ARROW_SIZE);
                    if (oNorthArrowSize != null)
                    {
                        double dNorthArrowSize = (double)oNorthArrowSize;
                        INorthArrow pNorthArrow = (INorthArrow)((IMapSurroundFrame)pElement).MapSurround;
                        pNorthArrow.Size = dNorthArrowSize;
                        ((IMapSurroundFrame)pElement).MapSurround = (IMapSurround)pNorthArrow;
                    }
                }
            }
        }

        //Show a layout element
        private static void ShowLayoutElement(ref IElement pElement)
        {            
            ITransform2D pTrans2d = (ITransform2D)pElement.Geometry;
            pTrans2d.Move((HIDE_LOCATION_X_OFFSET * -1), (HIDE_LOCATION_Y_OFFSET * -1));
            pElement.Geometry = (IGeometry)pTrans2d; //Update the geometry of the tagged element.

            //If the element is a text element with a text background and anchor point, move the anchor point.
            if (pElement is ITextElement)
            {
                ITextElement pTextElement = (ITextElement)pElement;
                ITextSymbol pTextSymbol = pTextElement.Symbol;
                if (pTextSymbol is IFormattedTextSymbol)
                {
                    IFormattedTextSymbol pFormattedTextSymbol = (IFormattedTextSymbol)pTextSymbol;
                    ITextBackground pTextBackGround = pFormattedTextSymbol.Background;
                    if (pTextBackGround is ICallout)
                    {
                        ICallout pCallout = (ICallout)pTextBackGround;
                        ITransform2D pTransformAnchorPoint = (ITransform2D)pCallout.AnchorPoint;
                        pTransformAnchorPoint.Move((HIDE_LOCATION_X_OFFSET * -1), (HIDE_LOCATION_Y_OFFSET * -1));
                        //Update the anchor point, text background, text symbol, and text element.
                        pCallout.AnchorPoint = (IPoint)pTransformAnchorPoint;
                        pFormattedTextSymbol.Background = (ITextBackground)pCallout;
                        pTextElement.Symbol = (ITextSymbol)pFormattedTextSymbol;
                    }
                }
            }
        }

        //Hide a layout element
        private static void HideLayoutElement(ref IElement pElement)
        {            
            ITransform2D pTrans2d = (ITransform2D)pElement.Geometry;
            pTrans2d.Move(HIDE_LOCATION_X_OFFSET, HIDE_LOCATION_Y_OFFSET);
            pElement.Geometry = (IGeometry)pTrans2d; //Update the geometry of the tagged element.

            //If the element is a text element with a text background and anchor point, move the anchor point.
            if (pElement is ITextElement)
            {
                ITextElement pTextElement = (ITextElement)pElement;
                ITextSymbol pTextSymbol = pTextElement.Symbol;
                if (pTextSymbol is IFormattedTextSymbol)
                {
                    IFormattedTextSymbol pFormattedTextSymbol = (IFormattedTextSymbol)pTextSymbol;
                    ITextBackground pTextBackGround = pFormattedTextSymbol.Background;
                    if (pTextBackGround is ICallout)
                    {
                        ICallout pCallout = (ICallout)pTextBackGround;
                        ITransform2D pTransformAnchorPoint = (ITransform2D)pCallout.AnchorPoint;
                        pTransformAnchorPoint.Move(HIDE_LOCATION_X_OFFSET, HIDE_LOCATION_Y_OFFSET);
                        //Update the anchor point, text background, text symbol, and text element.
                        pCallout.AnchorPoint = (IPoint)pTransformAnchorPoint;
                        pFormattedTextSymbol.Background = (ITextBackground)pCallout;
                        pTextElement.Symbol = (ITextSymbol)pFormattedTextSymbol;
                    }
                }                
            }
        }

        //This method retrieves a propertset from a layout element or creates a new property set.
        private static IPropertySet2 GetPropertySet(IElementProperties pElemProps)
        {
            //Retrieve the property set or create a new property set.
            object pCustomProperty = pElemProps.CustomProperty;
            IPropertySet2 pPropertySet = null;
            if (pCustomProperty == null) pPropertySet = new PropertySetClass();
            else if (pCustomProperty is PropertySet) pPropertySet = (IPropertySet2)pCustomProperty;
            else MessageBox.Show("The selected element contains custom properties that cannot be modified.", "JambWow - Could not modify layout element tags.");
            return pPropertySet;
        }

        //This method returns the selected layout element. If zero or multiple layout elements are selected, it returns null.
        private static IElementProperties GetSelectedLayoutElement()
        {
            IMxDocument pDoc = GenUtil.GetMxDocument();
            IActiveView pLayout = (IActiveView) pDoc.PageLayout;

            IGraphicsContainerSelect pGraphicsContSel = (IGraphicsContainerSelect)pDoc.PageLayout;
            IGraphicsContainer pGraphicsCont = (IGraphicsContainer)pDoc.PageLayout;

            //Get an enumeration of selected layout elements.
            IEnumElement pEnumSelected = pGraphicsContSel.SelectedElements;
            pEnumSelected.Reset();

            //Loop through each of the selected layout elements
            IElementProperties pElemProps = null;
            IElementProperties pReturnValue = null;
            int pElemPropsCount = 0;
            while ((pElemProps = (IElementProperties)pEnumSelected.Next()) != null)
            {
                pReturnValue = pElemProps;
                pElemPropsCount++;                
            }
            
            if (pElemPropsCount == 1)
                return pReturnValue;
            else if (pElemPropsCount == 0)
            {
                MessageBox.Show("Please select a layout element.", "JambWow - Could not tag layout element.");
                return null;
            }
            else
            {
                MessageBox.Show("More than one layout element is selected.", "JambWow - Could not tag layout element.");
                return null;
            }
        }

        #endregion
    }
}
