﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Drawing;
using System.Runtime.InteropServices;
using ESRI.ArcGIS.Framework;
using ESRI.ArcGIS.ArcMapUI;
using ESRI.ArcGIS.Carto;
using ESRI.ArcGIS.Output;
using ESRI.ArcGIS.esriSystem;
using ESRI.ArcGIS.Geometry;
using ESRI.ArcGIS.Display;

namespace JAMBWOW
{
    public static class PrintUtil
    {
        public static void PrintPages(PrintParameters pPP)
        {
            //Get starting page
            int startingPageNum = DDP_Util.GetCurrentKnownPage();
            //Make a list of pages to print
            
            List<int> PageNumbers = new List<int>();
            if (pPP.PageRangeType == PageRangeUtil.PageRangeTypes.PageList) //Make a list of select pages
            {
                //Parse the page numbers string
                string sPageNumbers = pPP.PageList;
                PageNumbers = PageRangeUtil.ParsePageNumbers(sPageNumbers);
            }
            else //Make a list of all pages
            {
                for (int pageNum = 1; pageNum <= DDP_Util.GetPageCount(); pageNum++)
                    PageNumbers.Add(pageNum);
            }

            //Remove disabled pages from PageNumbers list
            if (!pPP.OutputDisabledPages)
            {
                PageNumbers = PageRangeUtil.RemoveDisabledPages(PageNumbers);
            }

            //If Page Numbers list is empty, warn the user and exit.
            if (PageNumbers.Count == 0)
            {
                MessageBox.Show("All of the selected pages are disabled. No pages will be printed.", "JambWow");
                return;
            }

            //Create a progress report object.
            ProgressReport pProgressReport = new ProgressReport();
            pProgressReport.NumberOfItems = PageNumbers.Count;
            pProgressReport.CurrentItemNumber = 0;

            //Display a progress window.
            ProgressWindow pProgressWindow = new ProgressWindow("Printing Page");
            pProgressWindow.Show();

            try
            {
                //Loop through and print each page.
                for (int iIndex = 0; iIndex < PageNumbers.Count; iIndex++)
                {
                    int pageNum = PageNumbers[iIndex];

                    //Update the progress window
                    pProgressReport.CurrentItemNumber = iIndex + 1;
                    bool Canceled = pProgressWindow.ProgressChanged(pProgressReport);
                    if (Canceled)
                    {
                        DDP_Util.GoToPage(startingPageNum); //Return to the starting page
                        pProgressWindow.Close(); //Close the progress Window.
                        MessageBox.Show("Print Canceled", "JambWow");
                        return;
                    }
                    System.Windows.Forms.Application.DoEvents(); //Allow the UI to refresh

                    //Navigate to the page
                    DDP_Util.GoToPage(pageNum);

                    //Update the progress window after page navigation finishes.                
                    bool Canceled2 = pProgressWindow.ProgressChanged(pProgressReport);
                    if (Canceled2)
                    {
                        DDP_Util.GoToPage(startingPageNum); //Return to the starting page
                        pProgressWindow.Close(); //Close the progress Window.
                        MessageBox.Show("Print Canceled", "JambWow");
                        return;
                    }
                    System.Windows.Forms.Application.DoEvents(); //Allow the UI to refresh

                    //Initialize the cancel tracker object
                    ITrackCancel pTrackCancel = new CancelTrackerClass();
                    pTrackCancel.CancelOnClick = false;
                    pTrackCancel.CancelOnKeyPress = false;
                    //Assign the cancel tracker to the progress window.
                    pProgressWindow.CancelTracker = pTrackCancel;

                    //Print the requested number of copies.
                    for (int i = 1; i <= pPP.NumberOfCopies; i++)
                    {
                        PrintUtil.Print(pTrackCancel);
                    }

                    //Remove the cancel tracker from the progress window.
                    pProgressWindow.CancelTracker = null;
                }
                DDP_Util.GoToPage(startingPageNum); //Return to the starting page
                pProgressWindow.Close(); //Close the progress Window.                
                MessageBox.Show("Finished Printing " + PageNumbers.Count + " Page(s)", "JambWow");                
            }
            catch (Exception pException)
            {
                GenUtil.ShowError(pException);
                pProgressWindow.Close();
                DDP_Util.GoToPage(startingPageNum); //Return to the starting page
                MessageBox.Show("Print Failed", "JambWow");
            }
        }

        private static void Print(ITrackCancel pTrackCancel)
        {         
            //Get the printer object
            IApplication pApp = GenUtil.GetAppRef();
            IMxApplication pMxApp = (IMxApplication)pApp;
            IPrinter pPrinter = pMxApp.Printer;

            //Get the page object (which manages the on-screen representation of the printer page).
            IPageLayout pLayout = GenUtil.GetMxDocument().PageLayout;
            IPage pPage = pLayout.Page;

            //Set the page to printer mapping to "crop" mode.
            pPage.PageToPrinterMapping = esriPageToPrinterMapping.esriPageMappingCrop;            

            //Setup to Print
            int iDPI = pPrinter.Resolution;
            IEnvelope pDeviceBounds = new EnvelopeClass();
            pPage.GetDeviceBounds(pPrinter, 1, 0, (short)iDPI, pDeviceBounds);
            tagRECT rectDeviceBounds = new tagRECT();
            rectDeviceBounds.left =  (int)Math.Round(pDeviceBounds.XMin, MidpointRounding.AwayFromZero);
            rectDeviceBounds.top = (int)Math.Round(pDeviceBounds.YMin, MidpointRounding.AwayFromZero);
            rectDeviceBounds.right = (int)Math.Round(pDeviceBounds.XMax, MidpointRounding.AwayFromZero);
            rectDeviceBounds.bottom = (int)Math.Round(pDeviceBounds.YMax, MidpointRounding.AwayFromZero);
            //The following block of code fixes a problem with plots being cut off:
            //For emf printers we have to remove the top and left unprintable area
            //from device coordinates so its origin is 0,0.
            if (pPrinter is IEmfPrinter) 
            {
                rectDeviceBounds.right = rectDeviceBounds.right - rectDeviceBounds.left;
                rectDeviceBounds.bottom = rectDeviceBounds.bottom - rectDeviceBounds.top;
                rectDeviceBounds.left = 0;
                rectDeviceBounds.top = 0;
            }
            IEnvelope pVisBounds = new EnvelopeClass();
            pPage.GetPageBounds(pPrinter, 1, 0, pVisBounds);
            double paperWidthInch;
            double paperHeightInch;
            pPrinter.QueryPaperSize(out paperWidthInch, out paperHeightInch);
            IEnvelope devFrameEnvelope = new EnvelopeClass();
            devFrameEnvelope.PutCoords(0, 0, paperWidthInch * iDPI, paperHeightInch * iDPI);
            //(End of "Setup to Print")

            //Print
            IActiveView pLayoutActiveView = GenUtil.GetLayout();
            int hdc = pPrinter.StartPrinting(devFrameEnvelope, 0);
            pLayoutActiveView.Output(hdc, iDPI, rectDeviceBounds, pVisBounds, pTrackCancel);
            pPrinter.FinishPrinting();
        }
    }
}
